import rdflib

# All SBOL objects will be created in the default namespace,
# unless otherwise specified
DEFAULT_NS = "http://examples.org/"

# The URIs defined here determine the appearance of serialized RDF/XML nodes.
# Change these URIs to change the appearance of an SBOL class or property name
SBOL_URI = "http://sbols.org/v2"  # < Namespace for the SBOL standard.
RDF_URI = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
PURL_URI = "http://purl.org/dc/terms/"
PROV_URI = "http://www.w3.org/ns/prov"
PROVO = "http://www.w3.org/ns/prov"
SYSBIO_URI = "http://sys-bio.org"
# PROVO = "https://www.w3.org/TR/prov-o/"
OM_URI = 'http://www.ontology-of-units-of-measure.org/resource/om-2/'

# rdf nodes used in SBOL
NODENAME_ABOUT = "rdf:about"
NODENAME_RESOURCE = "rdf:resource"

VERSION_STRING = "1"

# URIs for SBOL objects
SBOL_DOCUMENT = SBOL_URI + "#Document"
SBOL_IDENTIFIED = SBOL_URI + "#Identified"
SBOL_DOCUMENTED = SBOL_URI + "#Documented"
SBOL_TOP_LEVEL = SBOL_URI + "#TopLevel"
SBOL_GENERIC_TOP_LEVEL = SBOL_URI + "#GenericTopLevel"
SBOL_SEQUENCE_ANNOTATION = SBOL_URI + "#SequenceAnnotation"
SBOL_COMPONENT = SBOL_URI + "#Component"
SBOL_FUNCTIONAL_COMPONENT = SBOL_URI + "#FunctionalComponent"
SBOL_COMPONENT_DEFINITION = SBOL_URI + "#ComponentDefinition"
SBOL_SEQUENCE = SBOL_URI + "#Sequence"
SBOL_MODULE_DEFINITION = SBOL_URI + "#ModuleDefinition"
SBOL_MODULE = SBOL_URI + "#Module"
SBOL_MODEL = SBOL_URI + "#Model"
SBOL_MAPS_TO = SBOL_URI + "#MapsTo"
SBOL_INTERACTION = SBOL_URI + "#Interaction"
SBOL_PARTICIPATION = SBOL_URI + "#Participation"
SBOL_SEQUENCE_CONSTRAINT = SBOL_URI + "#SequenceConstraint"
SBOL_LOCATION = SBOL_URI + "#Location"
SBOL_RANGE = SBOL_URI + "#Range"
SBOL_CUT = SBOL_URI + "#Cut"
SBOL_COLLECTION = SBOL_URI + "#Collection"
SBOL_GENERIC_LOCATION = SBOL_URI + "#GenericLocation"
SBOL_VARIABLE_COMPONENT = SBOL_URI + "#VariableComponent"
SBOL_COMBINATORIAL_DERIVATION = SBOL_URI + "#CombinatorialDerivation"
SBOL_ATTACHMENT = SBOL_URI + "#Attachment"
SBOL_IMPLEMENTATION = SBOL_URI + "#Implementation"
SBOL_EXPERIMENT = SBOL_URI + "#Experiment"
SBOL_EXPERIMENTAL_DATA = SBOL_URI + "#ExperimentalData"
SBOL_MEASURE = SBOL_URI + "#Measure"
UNDEFINED = SBOL_URI + "#Undefined"

# URIs for SBOL Properties
SBOL_IDENTITY = SBOL_URI + "#identity"
SBOL_PERSISTENT_IDENTITY = SBOL_URI + "#persistentIdentity"
SBOL_VERSION = SBOL_URI + "#version"
SBOL_DISPLAY_ID = SBOL_URI + "#displayId"
SBOL_NAME = PURL_URI + "title"
SBOL_DESCRIPTION = PURL_URI + "description"
SBOL_TYPES = SBOL_URI + "#type"
SBOL_START = SBOL_URI + "#start"
SBOL_END = SBOL_URI + "#end"
SBOL_SEQUENCE_ANNOTATIONS = SBOL_URI + "#sequenceAnnotation"
SBOL_COMPONENTS = SBOL_URI + "#component"
SBOL_COMPONENT_PROPERTY = SBOL_URI + "#component"
SBOL_ROLES = SBOL_URI + "#role"
SBOL_ELEMENTS = SBOL_URI + "#elements"
SBOL_ENCODING = SBOL_URI + "#encoding"
SBOL_SEQUENCE_PROPERTY = SBOL_URI + "#sequence"
SBOL_WAS_DERIVED_FROM = PROV_URI + "#wasDerivedFrom"
SBOL_DEFINITION = SBOL_URI + "#definition"
SBOL_ACCESS = SBOL_URI + "#access"
SBOL_DIRECTION = SBOL_URI + "#direction"
SBOL_MODELS = SBOL_URI + "#model"
SBOL_MODULES = SBOL_URI + "#module"
SBOL_FUNCTIONAL_COMPONENTS = SBOL_URI + "#functionalComponent"
SBOL_INTERACTIONS = SBOL_URI + "#interaction"
SBOL_MAPS_TOS = SBOL_URI + "#mapsTo"
SBOL_PARTICIPATIONS = SBOL_URI + "#participation"
SBOL_PARTICIPANT = SBOL_URI + "#participant"
SBOL_LOCAL = SBOL_URI + "#local"
SBOL_REMOTE = SBOL_URI + "#remote"
SBOL_REFINEMENT = SBOL_URI + "#refinement"
SBOL_SOURCE = SBOL_URI + "#source"
SBOL_LANGUAGE = SBOL_URI + "#language"
SBOL_FRAMEWORK = SBOL_URI + "#framework"
SBOL_SEQUENCE_CONSTRAINTS = SBOL_URI + "#sequenceConstraint"
SBOL_SUBJECT = SBOL_URI + "#subject"
SBOL_OBJECT = SBOL_URI + "#object"
SBOL_RESTRICTION = SBOL_URI + "#restriction"
SBOL_ORIENTATION = SBOL_URI + "#orientation"
SBOL_LOCATIONS = SBOL_URI + "#location"
SBOL_SOURCE_LOCATIONS = SBOL_URI + "#sourceLocation"
SBOL_ROLE_INTEGRATION = SBOL_URI + "#roleIntegration"
SBOL_MEMBERS = SBOL_URI + "#member"
SBOL_AT = SBOL_URI + "#at"
SBOL_OPERATOR = SBOL_URI + "#operator"
SBOL_VARIABLE_COMPONENTS = SBOL_URI + "#variableComponent"
SBOL_VARIABLE = SBOL_URI + "#variable"
SBOL_VARIANTS = SBOL_URI + "#variants"
SBOL_VARIANT_COLLECTIONS = SBOL_URI + "#variantCollections"
SBOL_VARIANT_DERIVATIONS = SBOL_URI + "#variantDeriviations"
SBOL_STRATEGY = SBOL_URI + "#strategy"
SBOL_TEMPLATE = SBOL_URI + "#template"
SBOL_ATTACHMENTS = SBOL_URI + "#attachment"
SBOL_MEASUREMENTS = SBOL_URI + "#measure"
SBOL_UNIT = SBOL_URI + "#unit"
SBOL_VALUE = SBOL_URI + "#hasNumericalValue"

# SBOL internal ontologies
# Option for Component::access or FunctionalComponent::access property
SBOL_ACCESS_PRIVATE = SBOL_URI + "#private"
# Option for Component::access or FunctionalComponent::access property
SBOL_ACCESS_PUBLIC = SBOL_URI + "#public"
# Option for FunctionalComponent::access property
SBOL_DIRECTION_IN = SBOL_URI + "#in"
# Option for FunctionalComponent::direction property
SBOL_DIRECTION_OUT = SBOL_URI + "#out"
# Option for FunctionalComponent::direction property
SBOL_DIRECTION_IN_OUT = SBOL_URI + "#inout"
# Option for FunctionalComponent::direction property
SBOL_DIRECTION_NONE = SBOL_URI + "#none"
# Option for SequenceConstraint::restriction property
SBOL_RESTRICTION_PRECEDES = SBOL_URI + "#precedes"
# Option for SequenceConstraint::restriction property
SBOL_RESTRICTION_SAME_ORIENTATION_AS = SBOL_URI + "#sameOrientationAs"
# Option for SequenceConstraint::restriction property
SBOL_RESTRICTION_OPPOSITE_ORIENTATION_AS = SBOL_URI + "#oppositeOrientationAs"
# Option for Sequence::encoding property
SBOL_ENCODING_IUPAC = "http://www.chem.qmul.ac.uk/iubmb/misc/naseq.html"
# Option for Sequence::encoding property
SBOL_ENCODING_IUPAC_PROTEIN = "http://www.chem.qmul.ac.uk/iupac/AminoAcid/"
# Option for Sequence::encoding property
SBOL_ENCODING_SMILES = "http://www.opensmiles.org/opensmiles.html"
# Option for Location::orientation property
SBOL_ORIENTATION_INLINE = SBOL_URI + "#inline"
# Option for Location::orientation property
SBOL_ORIENTATION_REVERSE_COMPLEMENT = SBOL_URI + "#reverseComplement"
# Option for FunctionalComponent::refinement property
SBOL_REFINEMENT_USE_REMOTE = SBOL_URI + "#useRemote"
# Option for FunctionalComponent::refinement property
SBOL_REFINEMENT_USE_LOCAL = SBOL_URI + "#useLocal"
# Option for MapsTo::refinement property
SBOL_REFINEMENT_VERIFY_IDENTICAL = SBOL_URI + "#verifyIdentical"
# Option for MapsTo::refinement property
SBOL_REFINEMENT_MERGE = SBOL_URI + "#merge"
# Option for SequenceAnnotation::roleIntegration
# or Component::roleIntegration property
SBOL_ROLE_INTEGRATION_MERGE = SBOL_URI + "#mergeRoles"
# Option for SequenceAnnotation::roleIntegration
# or Component::roleIntegration property
SBOL_ROLE_INTEGRATION_OVERRIDE = SBOL_URI + "#overrideRoles"
# Option for Usage::roles or Association::roles
SBOL_DESIGN = SBOL_URI + "#design"
# Option for Usage::roles or Association::roles
SBOL_BUILD = SBOL_URI + "#build"
# Option for Usage::roles or Association::roles
SBOL_TEST = SBOL_URI + "#test"
# Option for Usage::roles or Association::roles
SBOL_LEARN = SBOL_URI + "#learn"

# PROVO ontology
PROVO_ACTIVITY = PROVO + "#Activity"
PROVO_USAGE = PROVO + "#Usage"
PROVO_ASSOCIATION = PROVO + "#Association"
PROVO_AGENT = PROVO + "#Agent"
PROVO_PLAN = PROVO + "#Plan"
PROVO_WAS_GENERATED_BY = PROVO + "#wasGeneratedBy"
PROVO_STARTED_AT_TIME = PROVO + "#startedAtTime"
PROVO_ENDED_AT_TIME = PROVO + "#endedAtTime"
PROVO_QUALIFIED_ASSOCIATION = PROVO + "#qualifiedAssociation"
PROVO_QUALIFIED_USAGE = PROVO + "#qualifiedUsage"
PROVO_WAS_INFORMED_BY = PROVO + "#wasInformedBy"
PROVO_HAD_PLAN = PROVO + "#hadPlan"
PROVO_HAD_ROLE = PROVO + "#hadRole"
PROVO_AGENT_PROPERTY = PROVO + "#agent"
PROVO_ENTITY = PROVO + "#entity"

# Systems Biology Ontology
# Interaction.types
# Namespace for Systems Biology Ontology (SBO) terms
SBO = "http://identifiers.org/biomodels.sbo/SBO:"
# An SBO term and possible value for an Interaction::type property
SBO_INTERACTION = SBO + "0000343"
# An SBO term and possible value for an Interaction::type property
SBO_INHIBITION = SBO + "0000169"
# An SBO term and possible value for an Interaction::type property
SBO_GENETIC_PRODUCTION = SBO + "0000589"
# An SBO term and possible value for an Interaction::type property
SBO_NONCOVALENT_BINDING = SBO + "0000177"
# An SBO term and possible value for an Interaction::type property
SBO_STIMULATION = SBO + "0000170"
# An SBO term and possible value for an Interaction::type property
SBO_DEGRADATION = SBO + "0000179"
# An SBO term and possible value for an Interaction::type property
SBO_CONTROL = SBO + "0000168"
# An SBO term and possible value for an Interaction::type property
SBO_BIOCHEMICAL_REACTION = SBO + "0000176"
# An SBO term and possible value for an Interaction::type property
SBO_STIMULATED = SBO + "0000643"
# An SBO term and possible value for an Interaction::type property
SBO_CONVERSION = SBO + "0000182"

# Participant.roles
# An SBO term and possible value for an Participant::role property
SBO_PROMOTER = SBO + "0000598"
# An SBO term and possible value for an Participant::role property
SBO_GENE = SBO + "0000243"
# An SBO term and possible value for an Participant::role property
SBO_INHIBITOR = SBO + "0000020"
# An SBO term and possible value for an Participant::role property
SBO_INHIBITED = SBO + "0000642"
# An SBO term and possible value for an Participant::role property
SBO_STIMULATOR = SBO + "0000459"
# An SBO term and possible value for an Participant::role property
SBO_REACTANT = SBO + "0000010"
# An SBO term and possible value for an Participant::role property
SBO_PRODUCT = SBO + "0000011"
# An SBO term and possible value for an Participant::role property
SBO_LIGAND = SBO + "0000280"
# An SBO term and possible value for an Participant::role property
SBO_NONCOVALENT_COMPLEX = SBO + "0000253"
# An SBO term and possible value for an Participant::role property
SBO_BINDING_SITE = SBO + "0000494"
SBO_SUBSTRATE = SBO + "0000015"
SBO_COFACTOR = SBO + "0000604"
SBO_SIDEPRODUCT = SBO + "0000603"
# SBO_PRODUCT = SBO + "0000011"
SBO_ENZYME = SBO + "0000014"

# URIs for common Sequence Ontology terms
# Namespace for Sequence Ontology (SO) terms
SO = "http://identifiers.org/so/SO:"
# An SO term and possible value for ComponentDefinition::role property
SO_MISC = SO + "0000001"
# An SO term and possible value for ComponentDefinition::role property
SO_GENE = SO + "0000704"
# An SO term and possible value for ComponentDefinition::role property
SO_PROMOTER = SO + "0000167"
# An SO term and possible value for ComponentDefinition::role property
SO_CDS = SO + "0000316"
# An SO term and possible value for ComponentDefinition::role property
SO_RBS = SO + "0000139"
# An SO term and possible value for ComponentDefinition::role property
SO_TERMINATOR = SO + "0000141"
# An SO term and possible value for ComponentDefinition::role property
SO_SGRNA = SO + "0001998"
# An SO term and possible value for ComponentDefinition::role property
SO_LINEAR = SO + "0000987"
# An SO term and possible value for ComponentDefinition::role property
SO_CIRCULAR = SO + "0000988"
# An SO term and possible value for ComponentDefinition::role property
SO_PLASMID = SO + "0000155"

# BioPAX is used to indicate macromolecular and molecular types
# DNA
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_DNA = "http://www.biopax.org/release/biopax-level3.owl#DnaRegion"
# RNA
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_RNA = "http://www.biopax.org/release/biopax-level3.owl#RnaRegion"
# PROTEIN
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_PROTEIN = "http://www.biopax.org/release/biopax-level3.owl#Protein"
# SMALL_MOLECULE
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_SMALL_MOLECULE = "http://www.biopax.org/release/biopax-level3.owl#SmallMolecule"
# COMPLEX
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_COMPLEX = "http://www.biopax.org/release/biopax-level3.owl#Complex"

# EDAM ontology is used for Model.languages
# An EDAM ontology term and option for Model::language
EDAM_SBML = "http://identifiers.org/edam/format_2585"
# An EDAM ontology term and option for Model::language
EDAM_CELLML = "http://identifiers.org/edam/format_3240"
# An EDAM ontology term and option for Model::language
EDAM_BIOPAX = "http://identifiers.org/edam/format_3156"

# Model.frameworks
# SBO term and option for Model::framework
SBO_CONTINUOUS = SBO + "0000062"
# SBO term and option for Model::framework
SBO_DISCRETE = SBO + "0000063"

# URIs for SBOL extension objects
SYSBIO_DESIGN = SYSBIO_URI + "#Design"
SYSBIO_BUILD = SYSBIO_URI + "#Build"
SYSBIO_TEST = SYSBIO_URI + "#Test"
SYSBIO_ANALYSIS = SYSBIO_URI + "#Analysis"
SYSBIO_SAMPLE_ROSTER = SYSBIO_URI + "#SampleRoster"

IGEM_URI = "http://wiki.synbiohub.org/wiki/Terms/igem"

OM_NS = rdflib.Namespace(OM_URI)
OM_MEASURE: str = OM_NS.Measure.toPython()
OM_HAS_UNIT: str = OM_NS.hasUnit.toPython()
OM_HAS_NUMERICAL_VALUE: str = OM_NS.hasNumericalValue.toPython()



import os
import unittest

import requests

import sbol2 as sbol
import sbol2

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
CRISPR_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'crispr_example.xml')

TEST_RESOURCE = 'https://synbiohub.utah.edu'

if 'SBH_USER' in os.environ:
    username = os.environ['SBH_USER']
else:
    username = None
if 'SBH_PASSWORD' in os.environ:
    password = os.environ['SBH_PASSWORD']
else:
    password = None


class TestPartShop(unittest.TestCase):
    def test_pull_00(self):
        # Based on tutorial: https://pysbol2.readthedocs.io/en/latest/repositories.html
        doc = sbol.Document()
        igem = sbol.PartShop('https://synbiohub.org')
        igem.pull('https://synbiohub.org/public/igem/BBa_R0010/1', doc)
        # print(doc)
        # for obj in doc:
        #     print(obj)
        self.assertEqual(3, len(doc))

    def test_pull_01(self):
        # Based on tutorial: https://pysbol2.readthedocs.io/en/latest/repositories.html
        doc = sbol.Document()
        igem = sbol.PartShop('https://synbiohub.org/public/igem')
        igem.pull('BBa_B0032', doc)
        igem.pull('BBa_E0040', doc)
        igem.pull('BBa_B0012', doc)
        # print(doc)
        # for obj in doc:
        #     print(obj)
        self.assertEqual(7, len(doc))

    def test_pull_02(self):
        # I don't know what this adds over test_pull_01
        # This is a replacement for a previous test whose part was no
        # longer available at SynBioHub
        doc = sbol.Document()
        ps = sbol.PartShop('https://synbiohub.org/public/bsu')
        ps.pull('SyBiOntKB2sbol', doc)
        # print(doc)
        # for obj in doc:
        #     print(obj)
        self.assertEqual(1, len(doc))

    def test_login(self):
        # NOTE: Add /login because login pages may be different
        # depending on what site you're accessing
        igem = sbol.PartShop('https://synbiohub.org')
        response = igem.login('johndoe@example.org', 'test')
        self.assertEqual(response.status_code, 200)

    def test_login_bad(self):
        igem = sbol.PartShop('https://synbiohub.org')
        try:
            igem.login('johndoe@example.org', 'test1')
            self.fail('SBOLError not raised')
        except sbol.SBOLError as sbol_error:
            self.assertEqual(sbol_error.error_code(),
                             sbol.SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST)

    @unittest.skipIf(password is None, "No password supplied")
    def test_submit_00(self):
        RESOURCE = 'https://synbiohub.utah.edu'
        doc = sbol.Document()
        doc.displayId = 'test_collection'
        doc.name = 'test collection'
        doc.description = 'a test collection automatically generated ' \
                          'by the SBOL client library'
        ps = sbol.PartShop(RESOURCE)
        ps.login(username, password)
        response = ps.submit(doc, overwrite=1)
        self.assertEqual(response.status_code, 200)

    def test_sparqlQuery_00(self):
        ps = sbol.PartShop('https://synbiohub.org')
        response = ps.login('johndoe', 'test')
        self.assertEqual(response.status_code, 200)
        query = '''
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX dcterms: <http://purl.org/dc/terms/>
PREFIX dc: <http://purl.org/dc/elements/1.1/>
PREFIX sbh: <http://wiki.synbiohub.org/wiki/Terms/synbiohub#>
PREFIX prov: <http://www.w3.org/ns/prov#>
PREFIX sbol: <http://sbols.org/v2#>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX purl: <http://purl.obolibrary.org/obo/>
SELECT ?p ?o
WHERE {
  <https://synbiohub.org/public/igem/BBa_K318030/1> ?p ?o
}
'''
        response = ps.sparqlQuery(query)
        # print(response.text)
        self.assertEqual(response.status_code, 200)

    def test_getURL(self):
        expected_url = 'https://example.org'
        partShop = sbol.PartShop(expected_url)
        self.assertEqual(partShop.getURL(), expected_url)

    def test_getUser(self):
        url = 'https://example.org'
        partShop = sbol.PartShop(url)
        self.assertTrue(hasattr(partShop, 'getUser'))
        self.assertEqual(partShop.getUser(), '')
        user = 'scott'
        password = 'tiger'
        # Depending on DNS we might get one of two errors. Detect both
        # and handle them gracefully. If neither of the expected
        # errors is raised, then fail the test.
        try:
            partShop.login(user, password)
        except requests.exceptions.ConnectionError:
            pass
        except sbol.sbolerror.SBOLError as err:
            self.assertEqual(err.error_code(),
                             sbol.SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST)
        self.assertEqual(partShop.getUser(), user)

    def test_getSpoofedURL(self):
        url = 'https://example.org'
        spoofed_url = 'https://synbiohub.org'
        partShop = sbol.PartShop(url, spoofed_url=spoofed_url)
        self.assertTrue(hasattr(partShop, 'getSpoofedURL'))
        self.assertEqual(partShop.getSpoofedURL(), spoofed_url)

    def test_spoof(self):
        url = 'https://example.org'
        part_shop = sbol.PartShop(url)
        self.assertEqual(part_shop.getSpoofedURL(), '')
        spoofed_url = 'https://synbiohub.org'
        part_shop.spoof(spoofed_url)
        self.assertEqual(part_shop.getSpoofedURL(), spoofed_url)

    @unittest.skipIf(password is None, "No password supplied")
    def test_submit(self):
        # This test is derived from an etl-to-synbiohub_pipeline test
        # case that was failing.
        RESOURCE = 'https://tang.ece.utah.edu'
        SPOOFED_RESOURCE = 'https://synbiohub.utah.edu'
        doc = sbol.Document()
        doc.version = '1'
        doc.displayId = 'sbol_test'
        doc.name = "SBOL Test Collection"
        doc.description = "A scratch collection for automated testing of the sbol."
        sbh = sbol.PartShop(RESOURCE, SPOOFED_RESOURCE)
        sbh.login(username, password)
        try:
            sbh.submit(doc)
        except Exception:
            # What exception type should we really be expecting? The
            # original test had a bare except.
            uri_template = '{0}/user/{1}/{2}/{2}_collection/1'
            target_collection = uri_template.format(RESOURCE, username, doc.displayId)
            sbh.submit(doc, target_collection, 1)

    def test_uri2url(self):
        # This tests an internal method that underlies `remove` and
        # `pull`. No spoofing in these tests.
        part_shop = sbol.PartShop('https://synbiohub.org')
        uri = 'https://synbiohub.org/design/CAT_C0378/1'
        url = part_shop._uri2url(uri)
        self.assertEqual(url, uri)
        uri = 'https://example.org/design/CAT_C0378/1'
        try:
            url = part_shop._uri2url(uri)
        except sbol.SBOLError as err:
            self.assertEqual(err.error_code(),
                             sbol.SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT)
        else:
            self.fail('Expected SBOLError')

    def test_uri2url_spoofed(self):
        # Test with a spoofed url
        #
        # Spoofed URLs had been an issue in `pull` before extracting
        # the `_uri2url` method
        part_shop = sbol.PartShop('https://synbiohub.org',
                                  spoofed_url='https://example.org')
        uri = 'https://synbiohub.org/design/CAT_C0378/1'
        url = part_shop._uri2url(uri)
        self.assertEqual(url, uri)
        # _uri2url should change the spoofed URI to the resource URL
        uri = 'https://example.org/design/CAT_C0378/1'
        expected = 'https://synbiohub.org/design/CAT_C0378/1'
        url = part_shop._uri2url(uri)
        self.assertEqual(url, expected)
        # URI is not in spoofed or resource domains
        uri = 'https://example.com/design/CAT_C0378/1'
        try:
            url = part_shop._uri2url(uri)
        except sbol.SBOLError as err:
            self.assertEqual(err.error_code(),
                             sbol.SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT)
        else:
            self.fail('Expected SBOLError')

    @unittest.skipIf(password is None, "No password supplied")
    def test_attach_file(self):
        doc = sbol2.Document()
        doc.displayId = 'test_attachment'
        doc.name = 'test attachment'
        desc = 'a test collection created by the sbol2 unit tests'
        doc.description = desc
        md = doc.moduleDefinitions.create('attachmd')
        sbh = sbol2.PartShop(TEST_RESOURCE)
        sbh.login(username, password)
        sbh.submit(doc, overwrite=1)
        md_uri = '{}/user/{}/{}/{}/{}'.format(sbh.getURL(), sbh.getUser(),
                                              doc.displayId, md.displayId,
                                              md.version)
        sbh.attachFile(md_uri, CRISPR_LOCATION)

    def test_search_general(self):
        sbh = sbol2.PartShop(TEST_RESOURCE)
        # sbh.login(username, password)
        results = sbh.search("NAND")
        # The response is a list
        self.assertEqual(list, type(results))
        # There are 25 items in the list (search returns more,
        # but by default we get the first 25)
        self.assertEqual(25, len(results))
        # The response items are all of type Identified
        self.assertTrue(all([isinstance(x, sbol2.Identified)
                             for x in results]))

    def test_search_exact(self):
        igem = sbol.PartShop('https://synbiohub.org')
        limit = 10
        results = igem.search(sbol2.SO_PROMOTER,
                              sbol2.SBOL_COMPONENT_DEFINITION,
                              sbol2.SBOL_ROLES,
                              0, limit)
        # The response is a list
        self.assertEqual(list, type(results))
        # The response contains _limit_ items
        self.assertEqual(limit, len(results))
        # The response items are all of type Identified
        self.assertTrue(all([isinstance(x, sbol2.Identified)
                             for x in results]))
        doc = sbol2.Document()
        igem.pull([x.identity for x in results], doc, False)
        self.assertEqual(10, len(doc))
        for cd in doc.componentDefinitions:
            self.assertIn(sbol2.SO_PROMOTER, cd.roles)

    def test_pull_doc_version(self):
        # After a pull the document version was erroneously set to None (#281)
        doc = sbol.Document()
        self.assertIsNotNone(doc.version)
        old_version = doc.version
        igem = sbol.PartShop('https://synbiohub.org')
        igem.pull('https://synbiohub.org/public/igem/BBa_R0010/1', doc)
        self.assertEqual(3, len(doc))
        self.assertIsNotNone(doc.version)
        self.assertEqual(old_version, doc.version)


if __name__ == '__main__':
    unittest.main()



import unittest

import rdflib

import sbol2 as sbol


class TestInteraction(unittest.TestCase):

    def __init__(self, *args):
        super().__init__(*args)
        self.homespace = 'http://example.org'

    def make_identity(self, homespace, name, version, tipe='Interaction'):
        uri = '{}/{}/{}/{}'
        return uri.format(homespace, tipe, name, version)

    def setUp(self):
        # Always reset homespace
        sbol.setHomespace(self.homespace)
        sbol.Config.setOption('sbol_typed_uris', True)
        sbol.Config.setOption('sbol_compliant_uris', True)

    def testEmptyConstructor(self):
        # This is the default name in the interaction constructor
        name = 'example'
        i = sbol.Interaction()
        expected_identity = self.make_identity(self.homespace, name,
                                               sbol.VERSION_STRING)
        self.assertEqual(i.identity, expected_identity)
        self.assertEqual(i.displayId, name)
        self.assertEqual(i.version, sbol.VERSION_STRING)
        self.assertEqual(i.rdf_type, sbol.SBOL_INTERACTION)
        md = sbol.ModuleDefinition()
        md.interactions.add(i)
        self.assertEqual(len(md.interactions), 1)

    def test1ArgConstructor(self):
        name = 'foo'
        i = sbol.Interaction(name)
        expected_identity = self.make_identity(self.homespace, name,
                                               sbol.VERSION_STRING)
        self.assertEqual(i.identity, expected_identity)
        self.assertEqual(i.displayId, name)
        self.assertEqual(i.version, sbol.VERSION_STRING)
        self.assertEqual(i.rdf_type, sbol.SBOL_INTERACTION)
        md = sbol.ModuleDefinition()
        md.interactions.add(i)
        self.assertEqual(len(md.interactions), 1)

    def test2ArgConstructor(self):
        name = 'foo'
        i = sbol.Interaction(name, sbol.SBO_INHIBITION)
        expected_identity = self.make_identity(self.homespace, name,
                                               sbol.VERSION_STRING)
        self.assertEqual(i.identity, expected_identity)
        self.assertEqual(i.displayId, name)
        self.assertEqual(i.types, [sbol.SBO_INHIBITION])
        self.assertEqual(i.rdf_type, sbol.SBOL_INTERACTION)
        md = sbol.ModuleDefinition()
        md.interactions.add(i)
        self.assertEqual(len(md.interactions), 1)

    def test3ArgConstructor(self):
        name = 'foo'
        rdf_type = sbol.SBOL_INTERACTION + '2'
        i = sbol.Interaction(name, sbol.SBO_INHIBITION, type_uri=rdf_type)
        expected_identity = self.make_identity(self.homespace, name,
                                               sbol.VERSION_STRING,
                                               'Interaction2')
        self.assertEqual(i.identity, expected_identity)
        self.assertEqual(i.displayId, name)
        self.assertEqual(i.types, [sbol.SBO_INHIBITION])
        self.assertEqual(i.rdf_type, rdf_type)
        # Verify that when added to a module definition, this
        # interaction is in the list of interactions despite having a
        # different rdf_type. That's because the rdf_type is not taken
        # into account when adding to the module definition's
        # interactions property.
        md = sbol.ModuleDefinition()
        md.interactions.add(i)
        self.assertEqual(len(md.interactions), 1)



import os
import unittest

import sbol2

MY_DIR = os.path.dirname(os.path.abspath(__file__))
PARTS_FILE = os.path.join(MY_DIR, 'resources', 'tutorial', 'parts.xml')
PROMOTER_URI = 'https://synbiohub.org/public/igem/BBa_R0040/1'

expected_sequence = '\
tccctatcagtgatagagattgacatccctatcagtgatagagatactgagcactactagagGCCATGCCATTGGCTTTTGATAGAGGAC\
AACTACTAGtactagATGAACATTAAAAATATAAATGCTAATGAGAAGATAATTGATAAAATTAAAACTTGTAATAATAATAAAGATATTAATCAAT\
GTTTATCTGAAATAGCAAAGATAATACATTGTGAATATTACCTATTCGCTATTATCTATCCTCACTCAATAATTAAACCTGATGTTTCAATTATAGA\
TAATTACCCTGAAAAATGGCGTAAATATTATGATGATGCCGGACTACTAGAATATGACCCTGTAGTCGATTACTCTAAGTCCCATCATTCACCAATT\
AATTGGAACGTATTCGAAAAAAAAACAATAAAAAAAGAGTCTCCGAATGTAATAAAAGAAGCACAGGAATCGGGACTCATTACTGGATTTAGCTTTC\
CAATTCATACTGCAAGTAATGGTTTTGGAATGCTCAGTTTTGCTCATTCAGATAAAGATATTTATACTGACAGTTTATTTTTACACGCTAGTACAAA\
TGTACCATTAATGCTTCCTTCTTTAGTCGATAATTATCAAAAAATAAATACGACACGTAAAAAGTCAGATTCTATTTTAACAAAAAGAGAAAAAGAA\
TGCTTAGCGTGGGCGAGTGAAGGAAAAAGTACATGGGATATTTCAAAAATACTTGGCTGCAGTGAGCGTACTGTCACTTTTCATTTAACCAATACTC\
AAATGAAACTCAATACAACTAACCGCTGCCAAAGTATTTCTAAAGCAATTTTAACTGGCGCCATTAATTGTCCATACCTTAAAAATTAAtactagag\
GTCAGTTTCACCTGTTTTACGTAAAAACCCGCTTCGGCGGGTTTTTACTTTTGG'


class TestSbolTutorial(unittest.TestCase):
    """Encode the SBOL tutorial in a unit test.

    See https://github.com/SynBioDex/Community-Media/blob/master/2019/IWBDA19/workshop/solution.ipynb
    """

    def setUp(self):
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, True)

    def init_tutorial(self):
        # Set the default namespace (e.g. "http://my_namespace.org")
        namespace = "http://my_namespace.org"
        homespace = sbol2.setHomespace(namespace)

        # Test homespace
        self.assertIsNone(homespace)
        self.assertEqual(sbol2.getHomespace(), namespace)

        # Create a new SBOL document
        doc = sbol2.Document()

        # Test empty document
        self.assertIsInstance(doc, sbol2.Document)
        self.assertEqual(len(doc), 0)

        return doc

    def get_device_from_xml(self, doc):
        # Load some generic parts from `parts.xml` into another Document
        generic_parts = sbol2.Document(PARTS_FILE)

        # Test loaded document
        self.assertEqual(len(generic_parts), 32)
        self.assertEqual(len(list(generic_parts.componentDefinitions)), 14)
        self.assertEqual(len(list(generic_parts.moduleDefinitions)), 0)

        # Copy the parts from `parts.xml` into your document.
        # Be sure to specify the original namespace `http://examples.org`
        generic_parts.copy("http://examples.org", doc)

        # Test copied document
        self.assertEqual(len(doc), 32)
        self.assertEqual(len(list(doc.componentDefinitions)), 14)
        self.assertEqual(len(list(doc.moduleDefinitions)), 0)

    def get_device_from_synbiohub(self, doc):
        # Start an interface to igem's public part shop on
        # SynBioHub. Located at `https://synbiohub.org/public/igem`
        partshop = sbol2.PartShop('https://synbiohub.org/public/igem')

        # Search the part shop for parts from the iGEM interlab study
        # using the search term `interlab`
        records = partshop.search('interlab')

        # Import the medium strength device into your document
        self.assertEqual(32, len(doc))
        partshop.pull(PROMOTER_URI, doc)
        self.assertEqual(35, len(doc))

    def extract_cds_from_devices(self, doc):
        # Extract the promoter from your document.
        self.promoter = doc.componentDefinitions[PROMOTER_URI]
        self.assertEqual(PROMOTER_URI, self.promoter.identity)

        # Extract the ribosomal binding site (rbs) `Q2` from your document.
        self.rbs = doc.componentDefinitions['Q2']
        self.assertEqual('http://my_namespace.org/ComponentDefinition/Q2/1',
                         self.rbs.identity)

        # Extract the coding region (cds) `LuxR` from your document.
        self.cds = doc.componentDefinitions['LuxR']
        self.assertEqual('http://my_namespace.org/ComponentDefinition/LuxR/1',
                         self.cds.identity)

        # Extract the terminator `ECK120010818` from your document.
        self.terminator = doc.componentDefinitions['ECK120010818']
        self.assertEqual('http://my_namespace.org/ComponentDefinition/ECK120010818/1',
                         self.terminator.identity)

    def create_new_device(self, doc: sbol2.Document):
        # Create a new empty device named `my_device`
        my_device = doc.componentDefinitions.create('my_device')

        # Assemble the new device from the promoter, rbs, cds, and terminator from above.
        my_device.assemblePrimaryStructure([self.promoter,
                                            self.rbs,
                                            self.cds,
                                            self.terminator],
                                           sbol2.IGEM_STANDARD_ASSEMBLY)
        # Expecting the 4 components above plus 3 assembly scars
        self.assertEqual(7, len(my_device.getPrimaryStructure()))

        # Set the role of the device with the Sequence Ontology term `gene`
        my_device.roles = [sbol2.SO_GENE]

        # Compile the sequence for the new device
        my_device.compile()
        self.assertEqual(expected_sequence, my_device.sequence.elements)

    def test_tutorial(self):
        doc = self.init_tutorial()
        self.get_device_from_xml(doc)
        self.get_device_from_synbiohub(doc)
        self.extract_cds_from_devices(doc)
        self.create_new_device(doc)



import collections
import os
import unittest
import warnings

import rdflib

import sbol2
import sbol2 as sbol

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
PARTS_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'tutorial',
                              'parts.xml')


class TestObject(unittest.TestCase):

    def test_getPropertyValues(self):
        d = sbol.Document()
        d.read(PARTS_LOCATION)
        cd = d.componentDefinitions['http://examples.org/ComponentDefinition/AmeR/1']
        values = cd.getPropertyValues(sbol.SBOL_NAME)
        self.assertEqual(len(values), 1)

    def test_getPropertyValue(self):
        d = sbol.Document()
        d.read(PARTS_LOCATION)
        cd = d.componentDefinitions['http://examples.org/ComponentDefinition/AmeR/1']
        expected = 'AmeR'
        name = cd.getPropertyValue(sbol.SBOL_NAME)
        self.assertEqual(name, expected)

    def test_compare(self):
        # This didn't work because the SBOLObject.__eq__ method was
        # broken.  See issue #62.
        sbol.setHomespace('http://example.org/Unit_Test')
        doc = sbol.Document()
        md1 = doc.moduleDefinitions.create('Foo1')
        self.assertEqual(len(doc.moduleDefinitions), 1)
        md2 = doc.moduleDefinitions.create('Foo2')
        self.assertEqual(len(doc.moduleDefinitions), 2)

    def test_eq(self):
        sbol.setHomespace('http://example.org/Unit_Test')
        doc = sbol.Document()
        md1 = doc.moduleDefinitions.create('Foo1')
        self.assertEqual(len(doc.moduleDefinitions), 1)
        md2 = sbol.ModuleDefinition(uri='Foo2')
        self.assertNotEqual(md1, md2)

    def test_type(self):
        cd = sbol.ComponentDefinition('foo')
        expected = sbol.SBOL_COMPONENT_DEFINITION
        self.assertEqual(cd.type, expected)
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('bar')
        expected = sbol.SBOL_MODULE_DEFINITION
        self.assertEqual(md.type, expected)

    def test_cast(self):
        cd = sbol.ComponentDefinition('foo')
        cd2 = cd.cast(sbol.ComponentDefinition)
        self.assertEqual(cd, cd2)
        with self.assertRaises(TypeError):
            cd.cast(sbol.ModuleDefinition)

    def test_set_property_value_deprecated(self):
        # Verify that setPropertyValue is deprecated
        md = sbol.ModuleDefinition('md')
        with warnings.catch_warnings(record=True) as warns:
            md.setPropertyValue('myprop', 'foo')
        self.assertEqual(len(warns), 1)

    def test_set_property_value(self):
        # Suppress the deprecation warnings
        with warnings.catch_warnings(record=True) as warns:
            self.deprecated_test_set_property_value()

    def deprecated_test_set_property_value(self):
        # The bottom line is that SBOLObject.getPropertyValue() always
        # returns a string. That's what pysbol did, and what sbol2
        # should do.
        my_property = 'http://example.com/myproperty'
        md = sbol.ModuleDefinition('md1')
        # value argument must be of type string. If not, expect type error
        with self.assertRaises(TypeError):
            md.setPropertyValue(my_property, 2)
        foo_str = 'foo'
        bar_str = 'bar'
        with self.assertRaises(TypeError):
            md.setPropertyValue(my_property, [foo_str, bar_str])
        # Test basic setting
        empty_str = ''
        foo_literal = rdflib.Literal(foo_str)
        md.setPropertyValue(my_property, foo_literal)
        self.assertEqual(md.getPropertyValue(my_property), foo_str)
        # Test overwriting a single existing value
        md.setPropertyValue(my_property, bar_str)
        self.assertEqual(md.getPropertyValue(my_property), bar_str)
        # Test setting an existing multi-value property
        md.properties[rdflib.URIRef(my_property)] = [foo_str, bar_str]
        self.assertEqual(md.getPropertyValue(my_property), foo_str)
        self.assertEqual(md.getPropertyValues(my_property), [foo_str, bar_str])
        baz_str = 'baz'
        md.setPropertyValue(my_property, baz_str)
        self.assertEqual(md.getPropertyValue(my_property), baz_str)
        self.assertEqual([baz_str, bar_str], md.getPropertyValues(my_property))
        # Unset the value
        md.setPropertyValue(my_property, empty_str)
        self.assertEqual(md.getPropertyValue(my_property), empty_str)
        # This may seem odd, but it is the way pySBOL/libSBOL worked,
        # so we do it for backward compatibility
        self.assertEqual([empty_str, bar_str], md.getPropertyValues(my_property))

        # What about a plain string? Does that get converted to URIRef or Literal?
        #  If a value is present, mimic that value's type
        #  Else make the value a Literal
        foo_uri = rdflib.URIRef(foo_str)
        md2 = sbol.ModuleDefinition('md2')
        # String gets converted to Literal
        md2.setPropertyValue(my_property, foo_str)
        self.assertEqual(md2.getPropertyValue(my_property), foo_str)
        # Ok to overwrite a Literal with a URIRef
        md2.setPropertyValue(my_property, foo_uri)
        self.assertEqual(md2.getPropertyValue(my_property), foo_str)
        # Now setting a str value should convert it to a URIRef
        # because that's what is already there
        md2.setPropertyValue(my_property, bar_str)
        self.assertEqual(md2.getPropertyValue(my_property), bar_str)
        # Ok to overwrite a URIRef with a Literal
        md2.setPropertyValue(my_property, foo_literal)
        self.assertEqual(md2.getPropertyValue(my_property), foo_str)
        # Unsetting the value
        md2.setPropertyValue(my_property, '')
        self.assertEqual(md2.getPropertyValue(my_property), empty_str)

    def test_serialize_property_value(self):
        # Set a property value to the empty string and verify that it
        # can be serialized and deserialized properly. The underlying
        # question is what does the internal representation need to be
        # to have that happen correctly.
        my_property = 'http://example.com/myproperty'
        empty_str = ''
        foo_str = 'foo'
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('md')
        with warnings.catch_warnings(record=True):
            md.setPropertyValue(my_property, foo_str)
        self.assertEqual(md.getPropertyValue(my_property), foo_str)
        with warnings.catch_warnings(record=True):
            md.setPropertyValue(my_property, empty_str)
        self.assertEqual(md.getPropertyValue(my_property), empty_str)
        # Now serialize the document to a string
        serialized = doc.writeString()
        doc2 = sbol.Document()
        doc2.readString(serialized)
        self.assertEqual(len(doc2.moduleDefinitions), 1)
        md2 = doc2.moduleDefinitions[0]
        self.assertEqual(md2.displayId, md.displayId)
        self.assertEqual(md2.getPropertyValue(my_property), empty_str)

    def test_this(self):
        # The `this` attribute should return the object itself. This
        # is for backward compatibility with pySBOL. It is a
        # deprecated attribute.
        md = sbol2.ModuleDefinition('md')
        # Enable all warnings
        warnings.simplefilter("default")
        with warnings.catch_warnings(record=True) as warns:
            self.assertEqual(md.this, md)
        # Verify that a warning was issued
        self.assertEqual(len(warns), 1)

    def test_compare_properties(self):
        # Test that two objects are equal even if a property is in a
        # different order
        expected1 = [sbol2.SBO_INHIBITION, sbol2.SBO_CONTROL]
        expected2 = [sbol2.SBO_CONTROL, sbol2.SBO_INHIBITION]
        md1 = sbol2.ModuleDefinition('md1')
        md1.roles = expected1
        md2 = sbol2.ModuleDefinition('md1')
        md2.roles = expected2
        self.assertTrue(md1.compare(md2))

    def test_compare_owned_objects(self):
        # Test that two objects are equal even if owned objects are in
        # a different order
        expected1 = [sbol2.Module('m1'), sbol2.Module('m2')]
        expected2 = [sbol2.Module('m2'), sbol2.Module('m1')]
        md1 = sbol2.ModuleDefinition('md1')
        md1.modules = expected1
        md2 = sbol2.ModuleDefinition('md1')
        md2.modules = expected2
        self.assertTrue(md1.compare(md2))

    def test_compare_recursive(self):
        # Test that the compare method is doing a recursive compare
        m1a = sbol2.Module('m1')
        m1b = sbol2.Module('m1')
        # m1a and m1b compare True
        self.assertTrue(m1a.compare(m1b))
        m2 = sbol2.Module('m2')
        # m2 is different from m1a and m1b
        self.assertFalse(m2.compare(m1a))
        self.assertFalse(m2.compare(m1b))

        md1a = sbol2.ModuleDefinition('md1')
        md1b = sbol2.ModuleDefinition('md1')
        # md1a and md1b compare True
        self.assertTrue(md1a.compare(md1b))
        md1a.modules = [m1a]
        md1b.modules = [m1b]
        # md1a and md1b still compare True
        self.assertTrue(md1a.compare(md1b))
        md1a.modules = [m2]
        # Now md1a and md1b compare False because of recursive
        # comparison of modules
        self.assertFalse(md1a.compare(md1b))

    def test_is_hashable(self):
        obj = sbol2.SBOLObject()
        self.assertTrue(isinstance(obj, collections.abc.Hashable))
        # Make sure the hash function returns a reasonable value. The
        # hashable test above can be fooled according to the
        # documentation.
        self.assertTrue(isinstance(hash(obj), int))

    def test_get_property(self):
        # If a property is unset, return an empty string instead of
        # an IndexError
        fc = sbol2.FunctionalComponent()
        self.assertEqual('', fc.getPropertyValue(sbol2.SBOL_NAME))

        # getPropertyValue always returns a string, even if the
        # property is an IntProperty
        c = sbol.Cut()
        self.assertEqual('0', c.getPropertyValue(sbol.SBOL_AT))


if __name__ == '__main__':
    unittest.main()



import time
from sbol2 import *

igem = PartShop('https://synbiohub.org/public/igem')
records = []
search_term = 'plasmid'
limit = 25
total_hits = igem.searchCount(search_term)
for offset in range(0, total_hits, limit):
    records.extend( igem.search(search_term, SBOL_COMPONENT_DEFINITION, offset, limit) )
    time.sleep(0.1)



from .identified import Identified
from .constants import *
from .property import IntProperty
from .property import OwnedObject
from .property import ReferencedObject
from .property import URIProperty
from rdflib import URIRef


class Location(Identified):
    """The Location class specifies the strand orientation of a Component."""
    def __init__(self, uri=URIRef('example'), orientation=SBOL_ORIENTATION_INLINE,
                 *, type_uri=SBOL_LOCATION, version=VERSION_STRING):
        super().__init__(type_uri=type_uri, uri=uri, version=version)
        self.orientation = URIProperty(self, SBOL_ORIENTATION,
                                       '1', '1', [], orientation)
        self.sequence = ReferencedObject(self, SBOL_SEQUENCE_PROPERTY,
                                         SBOL_SEQUENCE, '0', '1', [])


class Range(Location):
    """A Range object specifies a region via discrete,
    inclusive start and end positions that correspond to indices
    for characters in the elements String of a Sequence.
    Note that the index of the first location is 1,
    as is typical practice in biology, rather than 0,
    as is typical practice in computer science."""
    def __init__(self, uri=URIRef('example'), start=1, end=2,
                 *, type_uri=SBOL_RANGE, version=VERSION_STRING):
        super().__init__(uri=uri, type_uri=type_uri, version=version)
        self.start = IntProperty(self, SBOL_START, '0', '1', None, start)
        self.end = IntProperty(self, SBOL_END, '0', '1', None, end)

    def precedes(self, comparand):
        if self.end < comparand.start:
            return comparand.start - self.end
        else:
            return 0

    def follows(self, comparand):
        if self.start > comparand.end:
            return comparand.end - self.start
        else:
            return 0

    def adjoins(self, comparand):
        if comparand.end + 1 == self.start:
            return 1
        if self.end + 1 == comparand.start:
            return 1
        return 0

    def contains(self, comparand):
        if self.start <= comparand.start and self.end >= comparand.end:
            return comparand.length()
        else:
            return 0

    def overlaps(self, comparand):
        if self.start == comparand.start and self.end == comparand.end:
            return 0
        elif self.start < comparand.start and self.end < comparand.end \
                and self.end >= comparand.start:
            return self.end - comparand.start + 1
        elif self.start > comparand.start and self.end > comparand.end \
                and self.start <= comparand.end:
            return comparand.end - self.start + 1
        elif comparand.contains(self):
            return comparand.contains(self)
        else:
            return 0

    def length(self):
        return self.end + 1 - self.start


class Cut(Location):
    """The Cut class specifies a location between
    two coordinates of a Sequence's elements."""
    def __init__(self, uri=URIRef('example'), at=0,
                 *, type_uri=SBOL_CUT, version=VERSION_STRING):
        super().__init__(uri=uri, type_uri=type_uri, version=version)
        self.at = IntProperty(self, SBOL_AT, '1', '1', [], at)


class GenericLocation(Location):
    """the GenericLocation class is included as a starting point
    for specifying regions on Sequence objects with
    encoding properties other than IUPAC and potentially nonlinear structure.
    This class can also be used to set the orientation of a SequenceAnnotation
    and any associated Component when their parent ComponentDefinition is
    a partial design that lacks a Sequence."""
    def __init__(self, uri=URIRef('example'),
                 *, type_uri=SBOL_GENERIC_LOCATION, version=VERSION_STRING):
        super().__init__(uri=uri, type_uri=type_uri, version=version)


class OwnedLocation(OwnedObject):
    def __init__(self, property_owner, sbol_uri, lower_bound, upper_bound,
                 validation_rules=None, first_object=None):
        """Initialize a container and optionally put the first object in it.
        If validation rules are specified, they will be checked upon initialization.

        builder is a function that takes a single argument, a string,
        and constructs an object of appropriate type for this
        OwnedObject instance. For instance, if this OwnedObject is
        intended to hold ComponentDefinitions, then the builder should
        return an object of type ComponentDefinition.

        """
        super().__init__(property_owner, sbol_uri, Location, lower_bound, upper_bound,
                         validation_rules, first_object)

    def createRange(self, uri=URIRef('example')):
        return self.create(uri, Range)

    def createCut(self, uri=URIRef('example')):
        return self.create(uri, Cut)

    def createGenericLocation(self, uri=URIRef('example')):
        return self.create(uri, GenericLocation)

    def getRange(self, uri=''):
        range = self.get(uri)
        if isinstance(range, Range):
            return range
        raise TypeError('Found object is not of type Range')

    def getCut(self, uri=''):
        cut = self.get(uri)
        if isinstance(cut, Cut):
            return cut
        raise TypeError('Found object is not of type Cut')



import unittest

import sbol2
import sbol2 as sbol


class TestOwnedObject(unittest.TestCase):

    def setUp(self):
        # All tests in this file use the default homespace
        sbol.setHomespace('http://examples.org')
        # All tests in this file use SBOL compliant URIs
        sbol.Config.setOption(sbol.ConfigOptions.SBOL_COMPLIANT_URIS.value, True)
        sbol.Config.setOption(sbol.ConfigOptions.SBOL_TYPED_URIS.value, True)

    # Test all the classes that contain OwnedObjects to be sure they
    # provide a callable builder argument
    def test_containers(self):
        # Note: Alphabetical Order
        x = sbol.Activity()
        x = sbol.CombinatorialDerivation()
        x = sbol.Component()
        x = sbol.ComponentDefinition()
        x = sbol.Document()
        x = sbol.Interaction()
        x = sbol.Module()
        x = sbol.ModuleDefinition()
        x = sbol.Participation()
        x = sbol.SequenceAnnotation()

    # These are classes that are considered outside the core, and thus
    # outside the beta. They use OwnedObject, so they will need to be
    # updated in the future.
    @unittest.expectedFailure  # See issue #310
    def test_non_beta_classes(self):
        # These are all in dbtl.py
        x = sbol.Analysis()
        x = sbol.Build()
        x = sbol.Design()

    def test_create(self):
        doc = sbol.Document()
        thing = doc.moduleDefinitions.create('thing1')
        self.assertIs(type(thing), sbol.ModuleDefinition)
        # md = doc.moduleDefinitions.get(thing.identity)
        # self.assertEqual(thing.identity, md.identity)

    def test_create2(self):
        # Test that the create adds the new object to the owned object.
        #
        # This seems to be a problem in the implementation of
        # ModuleDefinition.assemble()
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('thing1')
        self.assertEqual(len(md.modules), 0)
        module = md.modules.create('thing2')
        self.assertEqual(len(md.modules), 1)

    def test_uri_generation(self):
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('foo')
        expected = 'http://examples.org/ModuleDefinition/foo/1'
        self.assertEqual(md.identity, expected)
        m = md.modules.create('bar')
        expected = 'http://examples.org/ModuleDefinition/foo/bar/1'
        self.assertEqual(m.identity, expected)

    def test_displayId_lookup(self):
        # Test lookup of an object by its display Id via the
        # __getitem__() method.
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('foo')
        m = md.modules.create('bar')
        m2 = md.modules['bar']
        self.assertEqual(m.identity, m2.identity)

    def test_uri_lookup(self):
        # Test lookup of an object by its URI via the
        # __getitem__() method.
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('foo')
        m = md.modules.create('bar')
        m2 = md.modules[m.identity]
        self.assertEqual(m.identity, m2.identity)

    def test_assignment(self):
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('foo')
        m1 = sbol.Module('m1')
        m2 = sbol.Module('m2')
        md.modules = [m1, m2]
        self.assertEqual(list(md.modules), [m1, m2])

    def test_value(self):
        doc = sbol2.Document()
        self.assertEqual([], doc.componentDefinitions.value)

    def test_set_no_doc(self):
        # Add a module when the parent module definition is not in a
        # Document
        md = sbol.ModuleDefinition('foo')
        m1 = sbol.Module('m1')
        md.modules.set([m1])
        self.assertEqual(m1, md.modules.get(0))
        self.assertEqual(m1, md.modules.find(m1.identity))

    def test_find(self):
        md = sbol.ModuleDefinition('foo')
        m1 = md.modules.create('m1')
        self.assertEqual(m1, md.modules.get(0))
        self.assertEqual(m1, md.modules.find(m1.identity))
        self.assertEqual(m1, md.modules.find(m1.displayId))
        self.assertEqual(m1, md.modules.find(str(m1.displayId)))
        self.assertEqual(m1, md.modules.find('m1'))

    def test_get_by_persistent_id(self):
        # This grows out of an issue with SYNBICT
        # Searching by persistent id was not working
        sbol2.Config.setHomespace('http://synbict.org')
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        doc = sbol2.Document()
        md = doc.moduleDefinitions.create('nand_circuit')
        # Search by the persistent identity
        search_string = md.persistentIdentity
        self.assertEqual(md, doc.getModuleDefinition(search_string))

    def test_remove_by_display_id(self):
        doc = sbol2.Document()
        cd = doc.componentDefinitions.create('cd1')
        self.assertIn(cd.displayId, doc.componentDefinitions)
        self.assertIn(cd.identity, doc.SBOLObjects)
        doc.componentDefinitions.remove(cd.displayId)
        self.assertNotIn(cd.displayId, doc.componentDefinitions)
        self.assertNotIn(cd.identity, doc.SBOLObjects)

    def test_remove_by_identity(self):
        doc = sbol2.Document()
        cd = doc.componentDefinitions.create('cd1')
        self.assertIn(cd.identity, doc.componentDefinitions)
        self.assertIn(cd.identity, doc.SBOLObjects)
        doc.componentDefinitions.remove(cd.identity)
        self.assertNotIn(cd.identity, doc.componentDefinitions)
        self.assertNotIn(cd.identity, doc.SBOLObjects)

    def test_remove_by_persistent_identity(self):
        doc = sbol2.Document()
        cd = doc.componentDefinitions.create('cd1')
        self.assertIn(cd.persistentIdentity, doc.componentDefinitions)
        self.assertIn(cd.identity, doc.SBOLObjects)
        doc.componentDefinitions.remove(cd.persistentIdentity)
        self.assertNotIn(cd.persistentIdentity, doc.componentDefinitions)
        self.assertNotIn(cd.identity, doc.SBOLObjects)

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.OwnedObject(obj, type_uri, int, '0', '*')
        self.assertIn(type_uri, obj.owned_objects)
        self.assertEqual([], obj.owned_objects[type_uri])
        self.assertEqual([], obj.thing.value)



import time
import timeit
import sbol2

EXAMPLE_PATH = '../test/SBOLTestSuite/SBOL2/RepressionModel.xml'
EXAMPLE_PATH = '../W303.xml'


def append_document(doc: sbol2.Document, path):
    doc.append(path)
    doc.clear()


doc = sbol2.Document()
start = time.time()
seconds = timeit.timeit(lambda: append_document(doc, EXAMPLE_PATH), number=1)
end = time.time()
print('timeit result: {}'.format(seconds))
print('elapsed: {}'.format(end-start))



from rdflib import URIRef

from .constants import *
from .identified import Identified
from .measurement import Measurement
from .property import URIProperty, ReferencedObject, OwnedObject


class Participation(Identified):

    def __init__(self, type_uri=SBOL_PARTICIPATION, uri=URIRef('example'),
                 participant='', version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        self.roles = URIProperty(self, SBOL_ROLES, '0', '*', [])
        self.participant = ReferencedObject(self, SBOL_PARTICIPANT,
                                            SBOL_FUNCTIONAL_COMPONENT,
                                            '1', '1', [], participant)
        self.measurements = OwnedObject(self, SBOL_MEASUREMENTS,
                                        Measurement, '0', '*', [])

    def addRole(self, new_role):
        val = self.roles
        val.append(new_role)
        self.roles = val

    def removeRole(self, index=0):
        val = self.roles
        del val[index]
        self.roles = val

    def define(self, species, role=""):
        raise NotImplementedError('Not yet implemented')



from rdflib import URIRef

from .constants import *
from .identified import Identified
from .location import OwnedLocation
from .mapsto import MapsTo
from .measurement import Measurement
from .property import URIProperty, OwnedObject, ReferencedObject


class ComponentInstance(Identified):
    def __init__(self, rdf_type, uri, definition, access, version):
        super().__init__(rdf_type, uri, version)
        self.definition = ReferencedObject(self, SBOL_DEFINITION,
                                           SBOL_COMPONENT_DEFINITION,
                                           '1', '1', [], definition)
        self.access = URIProperty(self, SBOL_ACCESS, '0', '1', [], access)
        self.mapsTos = OwnedObject(self, SBOL_MAPS_TOS, MapsTo,
                                   '0', '*', [])
        self.measurements = OwnedObject(self, SBOL_MEASUREMENTS, Measurement,
                                        '0', '*', [])


class Component(ComponentInstance):
    def __init__(self, uri=URIRef('example'), definition='',
                 access=SBOL_ACCESS_PUBLIC, version=VERSION_STRING):
        super().__init__(SBOL_COMPONENT, uri, definition, access, version)
        self.roles = URIProperty(self, SBOL_ROLES, '0', '*',
                                 [Component._role_set_role_integration])
        self.roleIntegration = URIProperty(self, SBOL_ROLE_INTEGRATION,
                                           '0', '1', [])
        self.sourceLocations = OwnedLocation(self, SBOL_SOURCE_LOCATIONS, '0', '*')

    def addRole(self, new_role):
        val = self.roles
        val.append(new_role)
        self.roles = val

    def removeRole(self, index=0):
        val = self.roles
        del val[index]
        self.roles = val

    @staticmethod
    def _role_set_role_integration(sbol_obj, arg):
        """SBOL 2.3.0 says that if a Component has roles then it
        MUST specify a roleIntegration.
        """
        if not isinstance(sbol_obj, Component):
            raise TypeError('{!r} is not of type Component'.format(sbol_obj))
        if arg:
            if not sbol_obj.roleIntegration:
                # If roles are specified and no roleIntegration is present
                # default to mergeRoles
                sbol_obj.roleIntegration = SBOL_ROLE_INTEGRATION_MERGE


class FunctionalComponent(ComponentInstance):
    def __init__(self, uri=URIRef('example'), definition='',
                 access=SBOL_ACCESS_PUBLIC, direction=SBOL_DIRECTION_NONE,
                 version=VERSION_STRING):
        super().__init__(SBOL_FUNCTIONAL_COMPONENT, uri,
                         definition, access, version)
        self.direction = URIProperty(self, SBOL_DIRECTION,
                                     '1', '1', [], direction)

    def connect(self, interface_component):
        raise NotImplementedError("Not yet implemented")

    def mask(self, masked_component):
        raise NotImplementedError("Not yet implemented")

    def override(selfself, masked_component):
        raise NotImplementedError("Not yet implemented")

    def isMasked(self):
        raise NotImplementedError("Not yet implemented")



import datetime
import unittest

import dateutil.parser

import sbol2
from sbol2 import Activity


class TestProvo(unittest.TestCase):

    def test_activity_agent(self):
        # Verify that the agent attribute is a ReferencedObject
        doc = sbol2.Document()
        agent = doc.agents.create('007')
        activity = doc.activities.create('spying')
        activity.agent = agent
        self.assertEqual(activity.agent, agent.identity)


class TestAssociation(unittest.TestCase):

    def test_add_remove_role(self):
        a = sbol2.Association('a1')
        self.assertEqual([], a.roles)
        a.addRole(sbol2.SO_PROMOTER)
        self.assertEqual([sbol2.SO_PROMOTER], a.roles)
        a.addRole(sbol2.SO_MISC)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC], a.roles)
        a.addRole(sbol2.SO_CDS)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC, sbol2.SO_CDS], a.roles)
        a.removeRole(1)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_CDS], a.roles)


class TestUsage(unittest.TestCase):

    def test_add_remove_role(self):
        usage = sbol2.Usage('a1')
        self.assertEqual([], usage.roles)
        usage.addRole(sbol2.SO_PROMOTER)
        self.assertEqual([sbol2.SO_PROMOTER], usage.roles)
        usage.addRole(sbol2.SO_MISC)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC], usage.roles)
        usage.addRole(sbol2.SO_CDS)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC, sbol2.SO_CDS], usage.roles)
        usage.removeRole(1)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_CDS], usage.roles)


class TestActivity(unittest.TestCase):

    def test_started_at_time(self):
        activity = Activity('activity1')
        self.assertIsNone(activity.startedAtTime)
        # Test setting via datetime.datetime
        my_start = datetime.datetime.now()
        activity.startedAtTime = my_start
        self.assertEqual(my_start, activity.startedAtTime)
        # Test setting via string
        start_string = '2016-03-16T20:12:00Z'
        activity.startedAtTime = start_string
        start_dt = dateutil.parser.parse(start_string)
        self.assertEqual(start_dt, activity.startedAtTime)

    def test_ended_at_time(self):
        activity = Activity('activity2')
        self.assertIsNone(activity.endedAtTime)
        # Test setting via datetime.datetime
        my_end = datetime.datetime.now()
        activity.endedAtTime = my_end
        self.assertEqual(my_end, activity.endedAtTime)
        # Test setting via string
        end_string = '2016-03-16T20:12:00Z'
        activity.endedAtTime = end_string
        end_dt = dateutil.parser.parse(end_string)
        self.assertEqual(end_dt, activity.endedAtTime)



import os
import unittest

import rdflib

import sbol2 as sbol

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
PARTS_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'tutorial',
                              'parts.xml')
CRISPR_LOCATION = os.path.join(MODULE_LOCATION, 'resources',
                               'crispr_example.xml')


class TestIdentified(unittest.TestCase):

    def test_getDisplayId_SBOLCompliant(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        expected = 'CRISPR_Template'
        crispr_template = sbol.ModuleDefinition(expected)
        self.assertEqual(crispr_template.displayId, expected)

    def test_getPersistentIdentity_SBOLCompliant(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        expected = 'http://sbols.org/CRISPR_Example/CRISPR_Template'
        self.assertEqual(crispr_template.persistentIdentity,
                         expected)

    def test_getVersion_SBOLCompliant(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        self.assertEqual(crispr_template.version, '1')

    def test_setDisplayId_SBOLCompliant(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        crispr_template.displayId = 'test'
        self.assertEqual(str(crispr_template.displayId), 'test')
        self.assertEqual(crispr_template.displayId, 'test')

    def test_setPersistentIdentity_SBOLCompliant(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        expected = 'test'
        crispr_template.persistentIdentity = expected
        self.assertEqual(crispr_template.persistentIdentity, expected)

    def test_setVersion_SBOLCompliant(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        crispr_template.version = '2'
        self.assertEqual(crispr_template.version, '2')

    def test_getDisplayId_hasHomespace(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', False)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        self.assertEqual(None, crispr_template.displayId)

    def test_getPersistentIdentity_hasHomespace(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', False)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        expected = 'http://sbols.org/CRISPR_Example/CRISPR_Template'
        self.assertEqual(crispr_template.persistentIdentity, expected)

    def test_getVersion_hasHomespace(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', False)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        self.assertEqual(crispr_template.version, '1')

    def test_setPersistentIdentity_hasHomespace(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', False)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        expected = 'test'
        crispr_template.persistentIdentity = expected
        self.assertEqual(crispr_template.persistentIdentity, expected)

    def test_setVersion_hasHomespace(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption('sbol_compliant_uris', False)
        sbol.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        crispr_template.version = '2'
        self.assertEqual(crispr_template.version, '2')

    def test_name(self):
        d = sbol.Document()
        d.read(PARTS_LOCATION)
        cd = d.componentDefinitions['http://examples.org/ComponentDefinition/AmeR/1']
        expected_name = 'AmeR'
        self.assertEqual(cd.name, expected_name)

    def test_was_derived_from(self):
        d = sbol.Document()
        d.read(PARTS_LOCATION)
        cd = d.componentDefinitions['http://examples.org/ComponentDefinition/AmeR/1']
        uri = 'https://synbiohub.programmingbiology.org/public/Cello_Parts/AmeR/1'
        expected = [uri]
        self.assertEqual(cd.wasDerivedFrom, expected)

    def test_was_derived_from2(self):
        d = sbol.Document()
        d.read(CRISPR_LOCATION)
        cd = d.componentDefinitions['http://sbols.org/CRISPR_Example/CRP_b/1.0.0']
        expected = []
        self.assertEqual(cd.wasDerivedFrom, expected)

    def test_was_generated_by(self):
        d = sbol.Document()
        d.read(PARTS_LOCATION)
        cd = d.componentDefinitions['http://examples.org/ComponentDefinition/AmeR/1']
        uri = 'http://examples.org/Activity/CelloUCF2sbol_Activity/1'
        expected = [uri]
        self.assertEqual(cd.wasGeneratedBy, expected)

    def test_was_generated_by2(self):
        d = sbol.Document()
        d.read(CRISPR_LOCATION)
        cd = d.componentDefinitions['http://sbols.org/CRISPR_Example/CRP_b/1.0.0']
        expected = []
        self.assertEqual(cd.wasGeneratedBy, expected)


class TestCopy(unittest.TestCase):

    def test_copy_properties(self):
        i = sbol.Identified('i')
        i.name = 'foo'
        i_copy = i.copy()
        self.assertEqual(i_copy.name, 'foo')

    def test_copy_child_objects(self):
        doc = sbol.Document()
        cd = doc.componentDefinitions.create('cd')
        c1 = cd.components.create('c1')
        c2 = cd.components.create('c2')
        sc1 = cd.sequenceConstraints.create('sc1')
        sc2 = cd.sequenceConstraints.create('sc2')
        sa1 = cd.sequenceConstraints.create('sa1')
        sa2 = cd.sequenceConstraints.create('sa2')

        doc2 = sbol.Document()
        cd_copy = cd.copy(target_doc=doc2)
        self.assertEqual([c.identity for c in cd.components],
                         [c.identity for c in cd_copy.components])
        self.assertEqual([sc.identity for sc in cd.sequenceConstraints],
                         [sc.identity for sc in cd_copy.sequenceConstraints])
        self.assertEqual([sa.identity for sa in cd.sequenceAnnotations],
                         [sa.identity for sa in cd_copy.sequenceAnnotations])

    def test_import_object_into_new_namespace(self):
        # When copying an object into a new namespace, confirm that it's URI is copied
        # into the new namespace. Also confirm that any ReferencedObject attributes
        # whose values point to an object in the old namespace are also copied into the
        # new namespace
        sbol.setHomespace('http://examples.org')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', False)
        doc = sbol.Document()
        comp = sbol.ComponentDefinition('cd')
        seq = sbol.Sequence('seq')
        doc.addComponentDefinition(comp)
        doc.addSequence(seq)
        comp.sequences = seq.identity

        # Import from old homespace into new homespace
        old_homespace = sbol.getHomespace()
        sbol.setHomespace('http://acme.com')
        comp_copy = comp.copy(None, old_homespace)

        # Verify new namespace was correctly substituted
        self.assertEqual(comp_copy.identity, 'http://acme.com/cd/1')
        self.assertEqual(comp_copy.persistentIdentity, 'http://acme.com/cd')
        self.assertEqual(comp_copy.sequences[0], 'http://acme.com/seq/1')

        # Verify wasDerivedFrom relationship
        self.assertEqual(comp_copy.wasDerivedFrom[0], comp.identity)

        # Ensure these are equal under the covers
        self.assertEqual(type(comp.properties[sbol.SBOL_SEQUENCE_PROPERTY][0]),
                         rdflib.URIRef)
        self.assertEqual(type(comp.properties[sbol.SBOL_SEQUENCE_PROPERTY][0]),
                         type(comp_copy.properties[sbol.SBOL_SEQUENCE_PROPERTY][0]))

    def test_replace_namespace(self):
        sbol.setHomespace('http://wallacecorporation.com')
        old_namespace = 'http://tyrellcorporation.com'
        old_uri = rdflib.URIRef(old_namespace + '/foo')
        new_uri = sbol.identified.replace_namespace(old_uri, old_namespace,
                                                    sbol.SBOL_COMPONENT_DEFINITION)
        self.assertEqual(type(new_uri), rdflib.URIRef)
        self.assertEqual(new_uri, rdflib.URIRef(sbol.getHomespace() + '/foo'))

    def test_import_into_nontyped_namespace_from_typed_namespace(self):
        # Copy an sbol-typed URI to a non-typed, sbol-compliant URI
        sbol.setHomespace('http://examples.org')
        sbol.Config.setOption('sbol_compliant_uris', True)
        sbol.Config.setOption('sbol_typed_uris', True)

        doc = sbol.Document()
        comp = sbol.ComponentDefinition('cd')
        seq = sbol.Sequence('seq')
        comp.sequences = seq.identity
        doc.addComponentDefinition(comp)
        doc.addSequence(seq)

        # Import the object into the new namespace
        sbol.Config.setOption('sbol_typed_uris', False)
        old_homespace = sbol.getHomespace()
        sbol.setHomespace('http://acme.com')
        comp_copy = comp.copy(None, old_homespace)

        # Verify new namespace was correctly substituted and type token was successfully
        # removed
        self.assertEqual(comp_copy.identity, 'http://acme.com/cd/1')
        self.assertEqual(comp_copy.persistentIdentity, 'http://acme.com/cd')
        self.assertEqual(comp_copy.sequences[0], 'http://acme.com/seq/1')

        # Verify wasDerivedFrom relationship
        self.assertEqual(comp_copy.wasDerivedFrom[0], comp.identity)

    def test_import_into_typed_namespace_from_nontyped_namespace(self):

        # Copy an sbol-typed URI to a non-typed, sbol-compliant URI
        sbol.setHomespace('http://examples.org')
        sbol.Config.setOption('sbol_typed_uris', False)
        sbol.Config.setOption('sbol_compliant_uris', True)

        doc = sbol.Document()
        comp = sbol.ComponentDefinition('cd')
        seq = sbol.Sequence('seq')
        comp.sequences = seq.identity
        doc.addComponentDefinition(comp)
        doc.addSequence(seq)

        # Import the object into the new namespace
        sbol.Config.setOption('sbol_typed_uris', True)
        old_homespace = sbol.getHomespace()
        sbol.setHomespace('http://acme.com')
        comp_copy = comp.copy(None, old_homespace)

        # Verify new namespace was correctly substituted and type token was successfully
        # added
        self.assertEqual(comp_copy.identity,
                         'http://acme.com/ComponentDefinition/cd/1')
        self.assertEqual(comp_copy.persistentIdentity,
                         'http://acme.com/ComponentDefinition/cd')
        self.assertEqual(comp_copy.sequences[0],
                         'http://acme.com/Sequence/seq/1')

        # Verify wasDerivedFrom relationship
        self.assertEqual(comp_copy.wasDerivedFrom[0], comp.identity)

    def test_copy_namespace(self):
        # Confirm that extension namespaces are copied into the target Document
        extension_prefix = 'extension_prefix'
        extension_namespace = 'http://examples.org#'
        doc = sbol.Document()
        target_doc = sbol.Document()
        cd = sbol.ComponentDefinition('cd')
        doc.addComponentDefinition(cd)
        doc.addNamespace(extension_namespace, extension_prefix)
        cd.extension_property = sbol.property.TextProperty(cd, extension_namespace +
                                                           'extension_property', '0',
                                                           '1', None, 'foo')
        cd_copy = cd.copy(target_doc)
        self.assertTrue(target_doc._namespaces[extension_prefix] ==
                        rdflib.URIRef(extension_namespace))

    def test_copy_and_increment_version(self):
        # When copying an object within the same Document, the version should be
        # automatically incrememented
        sbol.Config.setOption('sbol_typed_uris', False)
        doc = sbol.Document()
        comp = sbol.ComponentDefinition('foo', sbol.constants.BIOPAX_DNA, '1.0.0')
        doc.addComponentDefinition(comp)

        comp_copy = comp.copy()
        self.assertEqual(comp.version, '1.0.0')
        self.assertEqual(comp_copy.version, '2.0.0')
        self.assertEqual(comp_copy.identity, comp.persistentIdentity + '/2.0.0')
        self.assertEqual(comp_copy.wasDerivedFrom[0], comp.identity)
        self.assertEqual(comp_copy.types[0], sbol.constants.BIOPAX_DNA)

    def test_copy_to_new_document(self):
        sbol.Config.setOption('sbol_typed_uris', False)
        doc = sbol.Document()
        comp1 = sbol.ComponentDefinition('cd1', sbol.constants.BIOPAX_DNA, '2')
        comp2 = sbol.ComponentDefinition('cd2', sbol.constants.BIOPAX_DNA, '2')
        comp2.wasDerivedFrom = comp1.identity
        doc.addComponentDefinition([comp1, comp2])

        # Since the object is cloned, the wasDerivedFrom should not be a circular
        # reference (this would violate SBOL-spec validation rules)
        doc2 = sbol.Document()
        comp3 = comp2.copy(doc2)
        self.assertEqual(comp3.identity, comp2.identity)
        self.assertEqual(len(doc2.componentDefinitions), 1)

        # Since self was cloned, check that wasDerivedFrom is not a circular reference
        # (this would violate SBOL-spec validation rules)
        self.assertEqual(comp3.wasDerivedFrom[0], comp1.identity)
        self.assertNotEqual(comp3.wasDerivedFrom[0], comp2.identity)

        # Confirm version is the same as the copied object
        self.assertEqual(comp3.version, '2')


if __name__ == '__main__':
    unittest.main()



from .identified import Identified
from .toplevel import TopLevel
from .constants import *
from rdflib import URIRef
from .property import ReferencedObject, URIProperty, OwnedObject


class VariableComponent(Identified):

    def __init__(self, type_uri=SBOL_IDENTIFIED, uri=URIRef('example'),
                 repeat='http://sbols.org/v2#on', version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        self.variable = ReferencedObject(self, SBOL_VARIABLE,
                                         SBOL_COMPONENT, '0', '1', [])
        self.repeat = URIProperty(self, SBOL_OPERATOR, '1', '1', [], repeat)
        self.variants = ReferencedObject(self, SBOL_VARIANTS,
                                         SBOL_COMPONENT_DEFINITION,
                                         '0', '*', [])
        self.variantCollections = ReferencedObject(self,
                                                   SBOL_VARIANT_COLLECTIONS,
                                                   SBOL_COLLECTION,
                                                   '0', '*', [])
        self.variantDerivations = ReferencedObject(self,
                                                   SBOL_VARIANT_DERIVATIONS,
                                                   SBOL_COMBINATORIAL_DERIVATION,
                                                   '0', '*', [])


class CombinatorialDerivation(TopLevel):

    def __init__(self, type_uri=SBOL_COMBINATORIAL_DERIVATION,
                 uri=URIRef("example"),
                 strategy='http://sbols.org/v2#enumerate',
                 version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        # TODO in original source, it doesn't look like strategy is used
        self.strategy = URIProperty(self, SBOL_STRATEGY, '1', '1', [])
        self.masterTemplate = ReferencedObject(self, SBOL_TEMPLATE,
                                               SBOL_COMBINATORIAL_DERIVATION,
                                               '0', '1', [])
        self.variableComponents = OwnedObject(self, SBOL_VARIABLE_COMPONENTS,
                                              VariableComponent,
                                              '0', '*', [])



import sbol2

sbh = sbol2.PartShop('https://hub.sd2e.org')
sbh.login('sd2e')

doc = sbol2.Document()
uri = 'https://hub.sd2e.org/user/sd2e/design/MG1655_LPV3_LacI_Sensor_pTac_AmeR_Network/1'
sbh.pull(uri, doc, True)

print(len(doc.moduleDefinitions[uri].modules.value))



from rdflib import URIRef

from .toplevel import TopLevel
from .constants import *
from .property import URIProperty


class Implementation(TopLevel):

    def __init__(self, uri=URIRef("example"), version=VERSION_STRING,
                 *, type_uri=SBOL_IMPLEMENTATION):
        super().__init__(type_uri, uri, version)
        self.built = URIProperty(self, SBOL_URI+'#built', '0', '1', [])



from . import validation
from .component import FunctionalComponent
from .config import Config
from .constants import *
from .interaction import Interaction
from .module import Module
from .property import OwnedObject
from .property import ReferencedObject
from .property import URIProperty
from .toplevel import TopLevel

import rdflib


class ModuleDefinition(TopLevel):
    """ModuleDefinition is a principle top-level class for
    describing the function of a design."""

    # The roles property is an OPTIONAL set of URIs that clarifies
    # the intended function of a ModuleDefinition.
    # These URIs might identify descriptive biological roles,
    # such as "metabolic pathway" and "signaling cascade",
    # but they can also identify identify "logical" roles,
    # such as "inverter" or "AND gate", or other abstract roles
    # for describing the function of design.
    # Interpretation of the meaning of such roles currently depends
    # on the software tools that read and write them.
    _roles = None

    # The modules property is OPTIONAL and MAY specify a
    # set of Module objects contained by the ModuleDefinition.
    # While the ModuleDefinition class is analogous
    # to a specification sheet for a system of interacting biological
    # elements, the Module class represents the occurrence
    # of a particular subsystem within the system. Hence,
    # this class allows a system design to include multiple instances
    # of a subsystem, all defined by reference to the same ModuleDefinition.
    # For example, consider the ModuleDefinition for a network of two-input
    # repressor devices in which the particular repressors
    # have not been chosen yet. This ModuleDefinition could contain multiple
    # Module objects that refer to the same ModuleDefinition
    # of an abstract two-input repressor device. Note that the set
    # of relations between Module and ModuleDefinition objects
    # is strictly acyclic.
    modules = None

    # The interactions property is OPTIONAL and MAY specify a set of
    # Interaction objects within the ModuleDefinition.
    # The Interaction class provides an abstract, machine-readable
    # representation of entity behavior within a
    # ModuleDefinition. Each Interaction contains Participation objects
    # that indicate the roles of the FunctionalComponent objects
    # involved in the Interaction.
    interactions = None

    # The functionalComponents property is OPTIONAL and MAY specify a set of
    # FunctionalComponent objects contained by the ModuleDefinition.
    # Just as a Module represents an instance of a subsystem in the
    # overall system represented by a ModuleDefinition, a FunctionalComponent
    # represents an instance of a structural entity (represented by a
    # ComponentDefinition) in the system. This concept allows a
    # ModuleDefinition to assert different interactions
    # for separate copies of the same structural entity if needed.
    # For example, a ModuleDefinition might contain
    # multiple FunctionalComponent objects that refer to the same promoter
    # ComponentDefinition, but assert different interactions for these
    # promoter copies based on their separate positions in another
    # ComponentDefinition that represents the structure of the entire system.
    functionalComponents = None

    # The models property is OPTIONAL and MAY specify a set of
    # URI references to Model objects.
    # Model objects are placeholders that link ModuleDefinition objects
    # to computational models of any format.
    # A ModuleDefinition object can link to more than one Model since
    # each might encode system behavior in
    # a different way or at a different level of detail.
    models = None

    def __init__(self, uri=rdflib.URIRef("example"), version=VERSION_STRING,
                 sbol_type_uri=SBOL_MODULE_DEFINITION):
        """Construct a ModuleDefinition

        :param uri: A full URI including a scheme, namespace, and identifier.
        If SBOLCompliance configuration is enabled, then this argument
        is simply the displayId
        for the new object and a full URI will automatically be constructed.
        :param version: An arbitrary version string. If SBOLCompliance is enabled,
        this should be a valid [Maven version string](http://maven.apache.org/).
        :param sbol_type_uri: The RDF type for an extension class
        derived from this one (optional)
        """
        super().__init__(sbol_type_uri, uri, version)
        self.roles = URIProperty(self, SBOL_ROLES, '0', '*', None)
        self.models = ReferencedObject(self, SBOL_MODELS,
                                       SBOL_MODEL, '0', '*', [])
        self.functionalComponents = OwnedObject(self,
                                                SBOL_FUNCTIONAL_COMPONENTS,
                                                FunctionalComponent,
                                                '0', '*', [])
        self.modules = OwnedObject(self, SBOL_MODULES, Module,
                                   '0', '*', [])
        self.interactions = OwnedObject(self, SBOL_INTERACTIONS,
                                        Interaction,
                                        '0', '*', [validation.libsbol_rule_17])

    def addRole(self, new_role):
        val = self.roles
        val.append(new_role)
        self.roles = val

    def removeRole(self, index=0):
        val = self.roles
        del val[index]
        self.roles = val

    def setOutput(self, output):
        """Defines an output for a sub-Module.
        Useful for top-down assembly of Modules and sub-Modules.

        If a FunctionalComponent with the given definition does not exist yet,
        one will be autoconstructed. Otherwise the FunctionalComponent
        with the given definition will be inferred.
        Be warned that this inference may fail
        if there is more than one FunctionalComponent with the same definition.

        :param output: A ComponentDefinition that defines the output
        :return: A FunctionalComponent that is derived from the argument
        ComponentDefinition and configured as this ModuleDefinition's output
        (it's direction property is set to SBOL_DIRECTION_OUT)
        """
        raise NotImplementedError("Not yet implemented")

    def setOutputFunctional(self, output):
        """Configures a FunctionalComponent as an output for a Module.

        Useful for bottom-up assembly of Modules and sub-Modules.

        :param output: The FunctionalComponent that will be configured
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def setInput(self, input):
        """Defines an input for a sub-Module. Useful for top-down assembly
        of Modules and sub-Modules.

        If a FunctionalComponent with the given definition does not exist yet,
        one will be autoconstructed.
        Otherwise the FunctionalComponent with the given definition
        will be inferred. Be warned that this inference may fail
        if there is more than one FunctionalComponent with the same definition.

        :param input: A ComponentDefinition that defines the input.
        :return: A FunctionalComponent that is derived
        from the argument ComponentDefinition and configured
        as this ModuleDefinition's input
        (it's direction property is set to SBOL_DIRECTION_IN).
        """
        raise NotImplementedError("Not yet implemented")

    def setInputFunctional(self, input):
        """Configures a FunctionalComponent as an input for a Module.

        Useful for bottom-up assembly of Modules and sub-Modules.

        :param input: The FunctionalComponent that will be configured.
        :return: None
        """

    def connect(self, output, input):
        """Connects the output of a sub-Module with the input
        of another sub-Module. Auto-constructs MapsTo objects.

        :param output: A FunctionalComponent configured as a Module output
        (see setOutput).
        :param input: A FunctionalComponent configured as a Module input
        (see setInput).
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def override(self, highlevel, lowlevel):
        """Overrides a low-level component in an abstract sub-Module
        with a high-level component
        in a parent ModuleDefinition, for example when overriding
        an abstract template Module with explicit components.

        :param highlevel: A high-level FunctionalComponent
        :param lowlevel: A low-level FunctionalComponent in a nested sub-Module
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def applyToModuleHierarchy(self, callback=None, user_data=None):
        """Perform an operation on every ModuleDefinition in a
        structurally-linked hierarchy of ModuleDefinitions
        by applying a callback function. If no callback is specified,
        the default behavior is to return a list of each ModuleDefinition
        in the hierarchy.

        :param callback: A callback function
        :param user_data: Arbitrary user data which can be passed in and out
        of the callback as an argument.
        :return: A list of all ModuleDefinitions in the hierarchy.
        """
        if callback:
            callback(self, user_data)
        result = [self]
        if self.modules:
            # I have child modules
            for m in self.modules:
                md = self.doc.find(m.definition)
                result.extend(md.applyToModuleHierarchy(callback, user_data))
        return result

    def assemble(self, list_of_modules):
        """Assemble a high-level ModuleDefinition from lower-level submodules.

        Autoconstructs Module objects in the process.

        :param list_of_modules: A list of submodule ModuleDefinitions.
        :return: None
        """
        if not Config.getOption('sbol_compliant_uris'):
            msg = 'This method only works when SBOL-compliance is enabled'
            raise SBOLError(msg, SBOLError.SBOL_ERROR_COMPLIANCE)
        for mdef in list_of_modules:
            m = self.modules.create(mdef.displayId)
            m.definition = mdef.identity

    def getTypeURI(self):
        return SBOL_MODULE_DEFINITION



from .constants import *
from .identified import Identified
from rdflib import URIRef
from .property import URIProperty, ReferencedObject


class SequenceConstraint(Identified):

    def __init__(self, type_uri=SBOL_SEQUENCE_CONSTRAINT,
                 uri=URIRef('example'), subject='', obj='',
                 restriction=SBOL_RESTRICTION_PRECEDES,
                 version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        self.subject = ReferencedObject(self, SBOL_SUBJECT, SBOL_COMPONENT,
                                        '1', '1', [], subject)
        self.object = ReferencedObject(self, SBOL_OBJECT, SBOL_COMPONENT,
                                       '1', '1', [], obj)
        self.restriction = URIProperty(self, SBOL_RESTRICTION,
                                       '1', '1', [], restriction)



from .property import LiteralProperty
import packaging.version as pv
import rdflib


class VersionProperty(LiteralProperty):

    def convert_to_user(self, value):
        result = str(value)
        if result == '':
            # special case, empty strings are equivalent to None
            return None
        return result

    def convert_from_user(self, value):
        # Empty string is equivalent to None
        if value == '':
            value = None
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None and self.upper_bound == 1 and self.lower_bound == 0:
            return None
        try:
            version = pv.Version(value)
        except pv.InvalidVersion as e:
            raise ValueError(e)
        except TypeError as e:
            raise ValueError(e)
        return rdflib.Literal(str(version))

    @staticmethod
    def _make_version(major: int, minor: int, micro: int) -> pv.Version:
        return pv.Version(f'{major}.{minor}.{micro}')

    @staticmethod
    def increment_major(version: str) -> str:
        old = pv.Version(version)
        new = VersionProperty._make_version(old.major + 1, old.minor, old.micro)
        return str(new)



import unittest

import sbol2


class TestComponent(unittest.TestCase):

    def test_role(self):
        c = sbol2.Component('c1')
        self.assertTrue(hasattr(c, 'identity'))
        self.assertEqual(None, c.roleIntegration)
        self.assertEqual([], c.roles)
        c.roles = [sbol2.SO_PROMOTER]
        # The SBOL 2.3.0 spec says that if a component has roles then
        # it MUST have a roleIntegration. There is a validator that should
        # set roleIntegration if not set
        self.assertNotEqual(None, c.roleIntegration)
        self.assertEqual(sbol2.SBOL_ROLE_INTEGRATION_MERGE,
                         c.roleIntegration)
        c.roleIntegration = sbol2.SBOL_ROLE_INTEGRATION_OVERRIDE
        c.roles = []
        # It is ok to have a roleIntegration even if there are no roles
        self.assertEqual(sbol2.SBOL_ROLE_INTEGRATION_OVERRIDE,
                         c.roleIntegration)

    def test_add_remove_role(self):
        c = sbol2.Component('c1')
        self.assertEqual([], c.roles)
        c.addRole(sbol2.SO_PROMOTER)
        self.assertEqual([sbol2.SO_PROMOTER], c.roles)
        c.addRole(sbol2.SO_MISC)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC], c.roles)
        c.addRole(sbol2.SO_CDS)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC, sbol2.SO_CDS], c.roles)
        c.removeRole(1)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_CDS], c.roles)


if __name__ == '__main__':
    unittest.main()



import unittest
import warnings

import rdflib

import sbol2

CRISPR_URI = 'http://sbols.org/CRISPR_Example/CRISPR_Template'
CRISPR_URI_1 = 'http://sbols.org/CRISPR_Example/CRISPR_Template/1'
CRISPR_MD_URI_1 = 'http://sbols.org/CRISPR_Example/ModuleDefinition/CRISPR_Template/1'
CRISPR_CD_URI_1 = 'http://sbols.org/CRISPR_Example/ComponentDefinition/CRISPR_Template/1'


class TestConfig(unittest.TestCase):

    def test_setOption_invalidOption(self):
        with self.assertRaises(sbol2.SBOLError):
            sbol2.Config.setOption('invalid_option', 2)

    def test_setOption_invalidOptionVal(self):
        with self.assertRaises(sbol2.SBOLError):
            sbol2.Config.setOption(sbol2.ConfigOptions.VERBOSE.value, 'Yellow')

    def test_hasHomespace_True(self):
        sbol2.setHomespace('http://sbols.org/CRISPR_Example/')
        self.assertTrue(sbol2.hasHomespace())

    def test_hasHomespace_False(self):
        sbol2.setHomespace('')
        self.assertFalse(sbol2.hasHomespace())

    def test_openworld_noHomespace(self):
        """See: https://pysbol2.readthedocs.io/en/latest/getting_started.html"""
        sbol2.setHomespace('')
        sbol2.Config.setOption('sbol_compliant_uris', False)
        sbol2.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol2.ModuleDefinition(CRISPR_URI)
        self.assertEqual(CRISPR_URI, str(crispr_template))
        self.assertEqual(CRISPR_URI, crispr_template.identity)

    def test_openworld_useHomespace(self):
        """See: https://pysbol2.readthedocs.io/en/latest/getting_started.html"""
        # Note: In tutorial CRISPR_Example is followed by a trailing slash.
        # This results in two slashes instead of one in both the original pySBOL
        # and the new code.
        sbol2.setHomespace('http://sbols.org/CRISPR_Example')
        sbol2.Config.setOption('sbol_compliant_uris', False)
        sbol2.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol2.ModuleDefinition('CRISPR_Template')
        self.assertEqual(CRISPR_URI, str(crispr_template))
        self.assertEqual(CRISPR_URI, crispr_template.identity)

    def test_SBOLCompliant(self):
        """See: https://pysbol2.readthedocs.io/en/latest/getting_started.html"""
        sbol2.setHomespace('http://sbols.org/CRISPR_Example')
        sbol2.Config.setOption('sbol_compliant_uris', True)
        sbol2.Config.setOption('sbol_typed_uris', False)
        crispr_template = sbol2.ModuleDefinition('CRISPR_Template')
        self.assertEqual(CRISPR_URI_1, str(crispr_template))
        self.assertEqual(CRISPR_URI_1, crispr_template.identity)

    def test_SBOLCompliant_typed(self):
        """See: https://pysbol2.readthedocs.io/en/latest/getting_started.html"""
        sbol2.setHomespace('http://sbols.org/CRISPR_Example')
        sbol2.Config.setOption('sbol_compliant_uris', True)
        sbol2.Config.setOption('sbol_typed_uris', True)
        crispr_template_md = sbol2.ModuleDefinition('CRISPR_Template')
        expected_crispr_template_md = CRISPR_MD_URI_1
        self.assertEqual(expected_crispr_template_md, str(crispr_template_md))
        self.assertEqual(expected_crispr_template_md,
                         crispr_template_md.identity)
        crispr_template_cd = sbol2.ComponentDefinition('CRISPR_Template')
        expected_crispr_template_cd = CRISPR_CD_URI_1
        self.assertEqual(expected_crispr_template_cd, str(crispr_template_cd))
        self.assertEqual(expected_crispr_template_cd,
                         crispr_template_cd.identity)

    def test_ca_path(self):
        with warnings.catch_warnings(record=True) as warns:
            sbol2.Config.getOption('ca-path')
            sbol2.Config.setOption('ca-path', '/fake/path')
        self.assertEqual(len(warns), 2)


if __name__ == '__main__':
    unittest.main()



#   Copyright (C) 2018 pysbolgraph developers. All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
#   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.

from lxml import etree
from lxml.etree import tostring
from lxml.etree import QName

from rdflib.namespace import RDF
from rdflib import URIRef, Literal

rdfNS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
sbolNS = "http://sbols.org/v2#"


def is_ownership_relation(g, triple):
    subject = triple[0].toPython()
    predicate = triple[1].toPython()
    obj = triple[2]

    ownership_predicates = {
        sbolNS + 'module',
        sbolNS + 'mapsTo',
        sbolNS + 'interaction',
        sbolNS + 'participation',
        sbolNS + 'functionalComponent',
        sbolNS + 'sequenceConstraint',
        sbolNS + 'location',
        sbolNS + 'sourceLocation',
        sbolNS + 'sequenceAnnotation',
        sbolNS + 'measure'
    }

    if predicate in ownership_predicates:
        return True

    # SBOL2 reuses the "component" predicate as both an ownership predicate (in
    # the case of ComponentDefinition) and a referencing one (in the case of
    # SequenceAnnotation).
    #
    if predicate == sbolNS + 'component':
        if (triple[0], RDF.type, URIRef(sbolNS + 'SequenceAnnotation')) in g:
            return False
        else:
            return True

    return False


def ns_prefix_dict(g):
    """Return a dictionary of namespace, uri prefix pairs."""
    return {ns: prefix.toPython() for (ns, prefix) in g.namespaces()}


def serialize_sboll2(g):
    prefixes = ns_prefix_dict(g)
    prefixes['rdf'] = rdfNS
    prefixes['sbol'] = sbolNS

    subject_to_element = dict()

    owned_elements = set()

    for triple in g.triples((None, RDF.type, None)):
        subject = triple[0].toPython()
        the_type = triple[2].toPython()
        if subject in subject_to_element:
            etree.SubElement(subject_to_element[subject],
                             prefixify(RDF.type, prefixes, True),
                             attrib={
                                 QName(rdfNS, 'resource'): the_type
                             })
        else:
            subject_to_element[subject] = etree.Element(prefixify(the_type, prefixes,
                                                                  True),
                                                        attrib={
                                                            QName(rdfNS,
                                                                  'about'): subject
                                                        }
                                                        )

    for triple in g.triples((None, None, None)):
        subject = triple[0].toPython()
        predicate = triple[1].toPython()
        obj = triple[2]
        element = subject_to_element[subject]
        if predicate == RDF.type.toPython():
            continue
        if is_ownership_relation(g, triple):
            owned_element = subject_to_element[obj.toPython()]
            ownership_element = etree.SubElement(element,
                                                 prefixify(predicate, prefixes, True))
            ownership_element.append(owned_element)
            owned_elements.add(obj.toPython())
            continue
        if isinstance(obj, URIRef):
            etree.SubElement(element, prefixify(predicate, prefixes, True), attrib={
                QName(rdfNS, 'resource'): obj.toPython()
            })
        elif isinstance(obj, Literal):
            elem = etree.SubElement(element, prefixify(predicate, prefixes, True))
            elem.text = obj
        else:
            raise Exception()

    doc = etree.Element(QName(rdfNS, 'RDF'), nsmap=prefixes)

    for subject in subject_to_element:
        if subject not in owned_elements:
            element = subject_to_element[subject]
            doc.append(element)

    return tostring(doc, pretty_print=True)


def prefixify(iri, prefixes, create_new):
    for prefix in prefixes:
        prefix_iri = prefixes[prefix]
        if iri.startswith(prefix_iri):
            return QName(prefix_iri, iri[len(prefix_iri):])
    if not create_new:
        return iri
    fragment_start = iri.rfind('#')
    if fragment_start == -1:
        fragment_start = iri.rfind('/')
    if fragment_start == -1:
        return iri
    iri_prefix = iri[:fragment_start + 1]
    i = 0
    while True:
        prefix_name = 'ns' + str(i)
        if prefix_name not in prefixes:
            prefixes[prefix_name] = iri_prefix
            return QName(iri_prefix, iri[len(iri_prefix):])
        i = i + 1



import posixpath
from typing import Union

from rdflib import URIRef

from .object import SBOLObject
from .config import Config
from .config import ConfigOptions
from .config import getHomespace
from .config import hasHomespace
from .constants import *
from .property import ReferencedObject, TextProperty
from .property import URIProperty
from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode
from .config import parseClassName
from . import validation
from .versionproperty import VersionProperty


class Identified(SBOLObject):
    # The persistentIdentity property is OPTIONAL and has a data type of URI.
    # This URI serves to uniquely refer to a set of SBOL objects
    # that are different versions of each other.
    # An Identified object MUST be referred to
    # using either its identity URI or its persistentIdentity URI.
    # _persistentIdentity = None

    # The displayId property is an OPTIONAL identifier
    # with a data type of String.
    # This property is intended to be an intermediate
    # between name and identity that is machine-readable,
    # but more human-readable than the full URI
    # of an identity. If the displayId property is used,
    # then its String value SHOULD be locally unique
    # (global uniqueness is not necessary)
    # and MUST be composed of only alphanumeric or underscore characters
    # and MUST NOT begin with a digit.
    # _displayId = None

    # If the version property is used,
    # then it is RECOMMENDED that version numbering follow the conventions
    # of [semantic versioning](http://semver.org/),
    # particularly as implemented by [Maven](http://maven.apache.org/).
    # This convention represents versions as sequences
    # of numbers and qualifiers that are separated
    # by the characters '.' and '-' and are compared in lexicographical order
    # (for example, 1 < 1.3.1 < 2.0-beta).
    # For a full explanation, see the linked resources.
    # _version = None
    version: Union[str, VersionProperty]

    # The wasDerivedFrom property is OPTIONAL and has a data type of URI.
    # An SBOL object with this property refers to another SBOL object
    # or non-SBOL resource from which this object was derived.
    # If the wasDerivedFrom property of an SBOL object A
    # that refers to an SBOL object B has an identical persistentIdentity,
    # and both A and B have a version,
    # then the version of B MUST precede that of A.
    # In addition, an SBOL object MUST NOT refer to itself
    # via its own wasDerivedFrom property
    # or form a cyclical chain of references via its wasDerivedFrom property
    # and those of other SBOL objects. For example, the reference chain
    # "A was derived from B and B was derived from A" is cyclical.
    # _wasDerivedFrom = None

    # An Activity which generated this ComponentDefinition,
    # eg., a design process like codon-optimization
    # or a construction process like Gibson Assembly
    # _wasGeneratedBy = None

    # The name property is OPTIONAL and has a data type of
    # String. This property is intended to be displayed to a human
    # when visualizing an Identified object. If an Identified object
    # lacks a name, then software tools SHOULD instead display the
    # object's displayId or identity. It is RECOMMENDED that software
    # tools give users the ability to switch perspectives between name
    # properties that are human-readable and displayId properties that
    # are less human-readable, but are more likely to be unique.
    # _name = None

    # The description property is OPTIONAL and has a data type of String.
    # This property is intended to contain a more thorough
    # text description of an Identified object.
    # _description = None

    def __init__(self, type_uri=SBOL_IDENTIFIED, uri=URIRef('example'),
                 version=VERSION_STRING):
        super().__init__(type_uri, uri)
        self.persistentIdentity = URIProperty(self, SBOL_PERSISTENT_IDENTITY,
                                              '0', '1', None, URIRef(uri))
        self.displayId = TextProperty(self, SBOL_DISPLAY_ID, '0', '1',
                                      [validation.sbol_rule_10204])
        self.version = VersionProperty(self, SBOL_VERSION, '0', '1', None, version)
        self.name = TextProperty(self, SBOL_NAME, '0', '1', None)
        self.description = TextProperty(self, SBOL_DESCRIPTION, '0', '1', None)
        if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
            self.displayId = uri
            self.persistentIdentity = URIRef(posixpath.join(getHomespace(), uri))
            if Config.getOption(ConfigOptions.SBOL_TYPED_URIS.value) is True:
                if self.version:
                    self.identity = URIRef(posixpath.join(getHomespace(),
                                                          self.getClassName(type_uri),
                                                          uri, str(self.version)))
                else:
                    self.identity = URIRef(posixpath.join(getHomespace(),
                                                          self.getClassName(type_uri),
                                                          uri))
            else:
                if self.version:
                    self.identity = URIRef(posixpath.join(getHomespace(),
                                                          uri, str(self.version)))
                else:
                    self.identity = URIRef(posixpath.join(getHomespace(), uri))
        elif hasHomespace():
            self.identity = URIRef(posixpath.join(getHomespace(), uri))
            self.persistentIdentity = URIRef(posixpath.join(getHomespace(),
                                                            uri))
        # Provo hooks
        self.wasDerivedFrom = URIProperty(self, SBOL_WAS_DERIVED_FROM,
                                          '0', '*', None)
        self.wasGeneratedBy = ReferencedObject(self, PROVO_WAS_GENERATED_BY,
                                               PROVO_WAS_GENERATED_BY, '0', '*', [])
        # self._identity.validate() # TODO

    def generate(self):
        raise NotImplementedError("Not yet implemented")

    def update_uri(self):
        """
        Recursively generates SBOL compliant ids for an object and all
        its owned objects, then checks to make sure that these ids are unique.
        :return: None
        """
        if self.parent is None:
            raise Exception('update_uri: Parent cannot be None')
        parent = self.parent
        if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
            # Form compliant URI for child object
            persistent_id = parent.properties[SBOL_PERSISTENT_IDENTITY][0]
            persistent_id = posixpath.join(persistent_id, self.displayId)
            if len(parent.properties[SBOL_VERSION]) > 0:
                version = parent.properties[SBOL_VERSION][0]
            else:
                version = VERSION_STRING
            obj_id = posixpath.join(persistent_id, version)
            # Reset SBOLCompliant properties
            self.identity = obj_id
            self.persistentIdentity = persistent_id
            # Check for uniqueness of URI in local object properties
            matches = parent.find_property_value(SBOL_IDENTIFIED, obj_id)
            if len(matches) > 0:
                raise SBOLError("Cannot update SBOL-compliant URI. The URI " +
                                str(self.identity) + " is not unique",
                                SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE)
            for rdf_type, store in self.owned_objects.items():
                if rdf_type not in self._hidden_properties:
                    for nested_obj in store:
                        nested_obj.update_uri()
        # Check for uniqueness of URI in Document
        if parent.doc:
            matches = parent.doc.find_property_value(SBOL_IDENTITY, self.identity)
            if len(matches) > 0:
                raise SBOLError("Cannot update SBOL-compliant URI. "
                                "An object with URI " + str(self.identity) +
                                " is already in the Document",
                                SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE)

    def copy(self, target_doc=None, target_namespace=None, version=None):

        new_obj = self.__class__()

        # This namespace map will be used later when copying namespaces over to
        # the new Document
        if self.doc:
            namespace_map = {ns: p for p, ns in self.doc._namespaces.items()}

        # Copy properties
        for property_uri, value_store in self.properties.items():
            new_obj.properties[property_uri] = value_store.copy()

            # Add a non-default namespace to the target document if not present
            # (This can happen when copying extension properties not in the
            # SBOL namespace, for example.)
            if self.doc and target_doc is not None:
                property_namespace = URIRef(parseNamespace(property_uri))
                if property_namespace in namespace_map.keys():
                    prefix = namespace_map[property_namespace]
                    target_doc.addNamespace(property_namespace, prefix)

        # If caller specified a target_namespace argument, then import objects into this
        # new namespace. This involves replacing the target_namespace in ReferenceObject
        # URIs with the current Homespace. Don't overwrite namespaces for the
        # wasDerivedFrom field, which points back to the original object
        if target_namespace:

            # Map the identity of self into the target namespace
            if hasattr(self, 'identity'):
                old_uri = self.identity
                new_uri = replace_namespace(old_uri, target_namespace, self.getTypeURI())
                new_obj.identity = new_uri

            if hasattr(self, 'persistentIdentity'):
                old_uri = self.persistentIdentity
                new_uri = replace_namespace(old_uri, target_namespace, self.getTypeURI())
                new_obj.persistentIdentity = new_uri

            # Map any references to other SBOL objects in the Document into the new
            # namespace
            if self.doc is not None:

                # Collect ReferencedObject attributes
                reference_properties = [p for p in new_obj.__dict__.values() if
                                        isinstance(p, ReferencedObject)]

                # These URIProperty attributes should be treated like ReferencedObject
                # attributes
                if '_built' in new_obj.__dict__.keys():
                    reference_properties.append(new_obj.__dict__['_built'])

                for reference_property in reference_properties:
                    values = new_obj.properties[reference_property._rdf_type]
                    new_values = []
                    for uri in values:
                        if target_namespace in uri:

                            referenced_object = self.doc.find(uri)
                            if referenced_object is None:
                                continue
                            new_uri = replace_namespace(uri, target_namespace,
                                                        referenced_object.getTypeURI())
                            new_values.append(new_uri)
                    new_obj.properties[reference_property._rdf_type] = new_values

        # Set the new object's version according to the user specified parameter. If
        # user didnt't provide a version, then set it automatically based on self's
        # version (if it has one).
        if version:
            new_obj.version = version
        elif self.version is not None:

            # If user is copying into the same Document and namespace, then assume the
            # user is trying to create a new version of the object. The version should
            # be automatically incremented to avoid a URI collision with the original
            # object.  However, if user is copying into a different Document, then copy
            # the original object's version without incrementing
            if self.doc and not target_doc and not target_namespace:
                new_obj.version = VersionProperty.increment_major(self.version)
            else:
                new_obj.version = self.version

        # Now set up the identity based on the persistentIdentity and maybe version
        # In the case of a Document there is no persistentIdentity so skip this block
        if new_obj.persistentIdentity:
            if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS) and new_obj.version:
                new_obj.identity = posixpath.join(new_obj.persistentIdentity,
                                                  new_obj.version)
            else:
                new_obj.identity = new_obj.persistentIdentity

        # Assign the new object to the target Document
        if target_doc:
            target_doc.add(new_obj)
        elif self.doc:
            self.doc.add(new_obj)

        # When an object is simply being cloned, the value of wasDerivedFrom should be
        # copied exactly as is from self. However, when copy is being used to generate
        # a new entity, the wasDerivedFrom should point back to self.
        if self.identity == new_obj.identity:
            new_obj.wasDerivedFrom = self.wasDerivedFrom
        else:
            new_obj.wasDerivedFrom = self.identity

        # Copy child objects recursively
        for property_uri, object_list in self.owned_objects.items():
            # Don't copy hidden properties
            if target_doc and property_uri in self._hidden_properties:
                continue
            for o in object_list:
                o_copy = o.copy(target_doc, target_namespace, version)
                new_obj.owned_objects[property_uri].append(o_copy)
                o_copy.parent = self
                # o_copy.update_uri()

        return new_obj


def parseNamespace(uri):
    '''
    Utility function for parsing the namespace from an RDF type -- note, this
    will work on some SBOL object URIs, but not universally.  It is primarily
    for use with RDF types.
    '''
    rlimit = uri.rfind('#') + 1
    if rlimit:
        return uri[:rlimit]
    rlimit = uri.rfind('/') + 1
    if rlimit:
        return uri[:rlimit]
    return ''


def replace_namespace(old_uri, target_namespace, rdf_type):
    """
    Utility function for mapping an SBOL object's identity into a new namespace. The
    rdf_type is used to map to and from sbol-typed namespaces.
    """

    # Work around an issue where the Document itself is being copied and
    # doesn't have its own URI, so old_uri is None. Return empty string
    # because the identity is not allowed to be None.
    if old_uri is None:
        return ''

    # If the value is an SBOL-typed URI, replace both the namespace and class name
    class_name = parseClassName(rdf_type)
    replacement_target = target_namespace + '/' + class_name

    # If not an sbol typed URI, then just replace the namespace
    if replacement_target not in old_uri:
        replacement_target = target_namespace

    if Config.getOption('sbol_typed_uris'):
        # Map into a typed namespace
        replacement = getHomespace() + '/' + class_name
    else:
        # Map into a non-typed namespace
        replacement = getHomespace()

    new_uri = old_uri.replace(replacement_target, replacement)
    if type(old_uri) is URIRef:
        return URIRef(new_uri)
    return new_uri



import unittest

import sbol2


class TestError(unittest.TestCase):

    def test_sbol_error_warning(self):
        # There are lots of places where the SBOLError constructor is
        # called with arguments in the wrong order. A warning has been
        # added to note these occurrences until they can be fixed.
        with self.assertWarns(RuntimeWarning):
            sbol2.SBOLError(sbol2.SBOLErrorCode.NOT_FOUND_ERROR,
                            'Item not found')

    def test_what(self):
        msg = 'Item not found'
        error = sbol2.SBOLError(msg, sbol2.SBOLErrorCode.NOT_FOUND_ERROR)
        self.assertEqual(error.what(), msg)

    def test_error_code(self):
        msg = 'Item not found'
        code = sbol2.SBOLErrorCode.NOT_FOUND_ERROR
        error = sbol2.SBOLError(msg, code)
        self.assertEqual(error.error_code(), code)



import locale
import logging
import os
import unittest
import tempfile
import shutil
import rdflib
import rdflib.compare
import sbol2

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL_TEST_SUITE = os.path.join(MODULE_LOCATION, 'SBOLTestSuite')

DEBUG_ENV_VAR = 'SBOL_TEST_DEBUG'


class TestRoundTripSBOL2(unittest.TestCase):

    def setUp(self):
        # Create temp directory
        self.temp_out_dir = tempfile.mkdtemp()
        # Disable validation to avoid unnecessary round trips
        # to the online validator
        sbol2.Config.setOption(sbol2.ConfigOptions.VALIDATE, False)
        self.logger = logging.getLogger('sbol2.test')
        if not self.logger.hasHandlers():
            logging.basicConfig()
        if DEBUG_ENV_VAR in os.environ:
            self.logger.setLevel(logging.DEBUG)

    def tearDown(self):
        # Remove directory after the test
        if self.temp_out_dir:
            shutil.rmtree(self.temp_out_dir)
            self.temp_out_dir = None

    def run_round_trip_file(self, test_path):
        """Runs a round trip test on the file at the given path.
        Path can be relative or absolute.
        """
        filename = os.path.basename(test_path)
        test2_path = os.path.join(self.temp_out_dir, filename)
        # Read the document, then write it back to disk
        doc = sbol2.Document()
        doc.read(test_path)
        doc.write(test2_path)

        # Read the newly written document and compare results
        doc2 = sbol2.Document()
        doc2.read(test2_path)
        self.assertTrue(doc.compare(doc2))

        # Now compare the graphs in RDF
        g1 = rdflib.Graph()
        g1.load(test_path)
        iso1 = rdflib.compare.to_isomorphic(g1)
        g2 = rdflib.Graph()
        g2.load(test2_path)
        iso2 = rdflib.compare.to_isomorphic(g2)
        rdf_diff = rdflib.compare.graph_diff(iso1, iso2)
        if rdf_diff[1] or rdf_diff[2]:
            self.logger.warning('Detected %d different RDF triples in %s' %
                                (len(rdf_diff[1]) + len(rdf_diff[2]), test_path))
            if not self.logger.isEnabledFor(logging.DEBUG):
                self.logger.warning('Set environment variable %s to see details',
                                    DEBUG_ENV_VAR)
            for stmt in rdf_diff[1]:
                self.logger.debug('Only in original: %r', stmt)
            for stmt in rdf_diff[2]:
                self.logger.debug('Only in loaded: %r', stmt)
            self.fail('Differences in RDF detected')

    def run_round_trip_dir(self, directory, skip_list=None):
        dir_name = os.path.basename(directory)
        for filename in os.listdir(directory):
            if skip_list and filename in skip_list:
                continue
            if filename.endswith('rdf') or filename.endswith('xml'):
                test_path = os.path.join(directory, filename)
                with self.subTest(filename=os.path.join(dir_name, filename)):
                    self.setUp()
                    self.run_round_trip_file(test_path)
                    self.tearDown()

    def test_sbol2_files(self):
        test_dir = os.path.join(SBOL_TEST_SUITE, 'SBOL2')
        # test_source_location.xml has a serialization error in SBOL2Serialize
        self.run_round_trip_dir(test_dir)

    def test_sbol2_bp_files(self):
        test_dir = os.path.join(SBOL_TEST_SUITE, 'SBOL2_bp')
        self.run_round_trip_dir(test_dir)

    def test_sbol2_ic_files(self):
        test_dir = os.path.join(SBOL_TEST_SUITE, 'SBOL2_ic')
        self.run_round_trip_dir(test_dir)

    def test_sbol2_nc_files(self):
        test_dir = os.path.join(SBOL_TEST_SUITE, 'SBOL2_nc')
        # SBOL1and2Test.xml has a namespace error when serializing
        # Handle it for now as an expected failure below
        self.run_round_trip_dir(test_dir, ['SBOL1and2Test.xml'])

    @unittest.expectedFailure  # See issue #210
    def test_sbol_1_and_2(self):
        # Invalid namespace prefix in lxml.etree when serializing the document
        test_path = os.path.join(SBOL_TEST_SUITE, 'SBOL2_nc',
                                 'SBOL1and2Test.xml')
        self.run_round_trip_file(test_path)

    def test_utf8_roundtrip(self):
        # Test loading a utf-8 SBOL file without LANG set. This was a
        # bug at one time, and only shows itself when LANG is unset.
        # Here we simulate that by temporarily setting the locale to
        # the generic 'C' locale.
        test_path = os.path.join(SBOL_TEST_SUITE, 'SBOL2', 'pICSL50014.xml')
        loc = locale.getlocale()
        try:
            locale.setlocale(locale.LC_ALL, 'C')
            self.run_round_trip_file(test_path)
        finally:
            locale.setlocale(locale.LC_ALL, loc)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol2


class TestExperiment(unittest.TestCase):

    def test_exported(self):
        self.assertIn('Experiment', dir(sbol2))
        self.assertIn('ExperimentalData', dir(sbol2))


if __name__ == '__main__':
    unittest.main()



import unittest
import sbol2


class TestModuleDefinition(unittest.TestCase):

    def setUp(self):
        # All tests in this file use the default homespace
        sbol2.setHomespace('http://examples.org')
        # All tests in this file use SBOL compliant URIs
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS.value, True)
        sbol2.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS.value, True)

    # Borrowed from libSBOL/wrapper/unit_tests.py
    def testApplyCallbackRecursively(self):
        # Assemble module hierarchy
        doc = sbol2.Document()
        root = sbol2.ModuleDefinition('root')
        sub = sbol2.ModuleDefinition('sub')
        leaf = sbol2.ModuleDefinition('leaf')
        doc.addModuleDefinition([root, sub, leaf])
        root.assemble([sub])
        sub.assemble([leaf])

        # Define callback which performs an operation on the given ModuleDefinition
        def callback(md, params):
            level = params[0]
            level += 1
            params[0] = level

        # Apply callback
        level = 0
        params = [level]
        flattened_module_tree = root.applyToModuleHierarchy(callback, params)
        level = params[0]
        flattened_module_tree = [md.identity for md in flattened_module_tree]
        expected_module_tree = [md.identity for md in [root, sub, leaf]]
        self.assertSequenceEqual(flattened_module_tree, expected_module_tree)
        self.assertEqual(level, 3)

    def testAssemble(self):
        # Assemble module hierarchy
        doc = sbol2.Document()
        root = sbol2.ModuleDefinition('root')
        sub = sbol2.ModuleDefinition('sub')
        doc.addModuleDefinition([root, sub])
        root.assemble([sub])

    def testAttachments(self):
        # All TopLevels can have attachments
        # TODO: There's a better test here involving a Document
        attachment = sbol2.Attachment('attachment')
        md = sbol2.ModuleDefinition('md')
        md.attachments = attachment.identity
        self.assertEqual([attachment.identity], md.attachments)

    def test_add_remove_role(self):
        md = sbol2.ModuleDefinition('c1')
        self.assertEqual([], md.roles)
        md.addRole(sbol2.SO_PROMOTER)
        self.assertEqual([sbol2.SO_PROMOTER], md.roles)
        md.addRole(sbol2.SO_MISC)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC], md.roles)
        md.addRole(sbol2.SO_CDS)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC, sbol2.SO_CDS], md.roles)
        md.removeRole(1)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_CDS], md.roles)


if __name__ == '__main__':
    unittest.main()



from sbol2 import *

setHomespace('http://sys-bio.org')
doc = Document()

gene = ComponentDefinition('gene_example')
r0010 = ComponentDefinition('R0010')
b0032 = ComponentDefinition('B0032')
e0040 = ComponentDefinition('E0040')
b0012 = ComponentDefinition('B0012')

r0010.roles = SO_PROMOTER
b0032.roles = SO_CDS
e0040.roles = SO_RBS
b0012.roles = SO_TERMINATOR

doc.addComponentDefinition(gene)
doc.addComponentDefinition([ r0010, b0032, e0040, b0012 ])

gene.assemblePrimaryStructure([ r0010, b0032, e0040, b0012 ])

first = gene.getFirstComponent()
print(first.identity)
last = gene.getLastComponent()
print(last.identity)

r0010.sequence = Sequence('R0010', 'ggctgca')
b0032.sequence = Sequence('B0032', 'aattatataaa')
e0040.sequence = Sequence('E0040', "atgtaa")
b0012.sequence = Sequence('B0012', 'attcga')

target_sequence = gene.compile()
print(gene.sequence.elements)

result = doc.write('gene_cassette.xml')
print(result)



import unittest
import sbol2


class TestSequenceAnnotation(unittest.TestCase):

    def test_add_remove_role(self):
        sa = sbol2.SequenceAnnotation()
        self.assertEqual([], sa.roles)
        sa.addRole(sbol2.SO_PROMOTER)
        self.assertEqual([sbol2.SO_PROMOTER], sa.roles)
        sa.addRole(sbol2.SO_MISC)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC], sa.roles)
        sa.addRole(sbol2.SO_CDS)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC, sbol2.SO_CDS], sa.roles)
        sa.removeRole(1)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_CDS], sa.roles)

    def test_component(self):
        cd = sbol2.ComponentDefinition('cd')
        sa = cd.sequenceAnnotations.create('sa')
        c = cd.components.create('c')
        sa.component = c
        self.assertEqual(sa.component, c.identity)


if __name__ == '__main__':
    unittest.main()



from deprecated import deprecated
from rdflib import URIRef

from .constants import *
from .property import URIProperty, TextProperty
from .toplevel import TopLevel
from .config import Config
from .sbolerror import SBOLError, SBOLErrorCode
from .location import Range


class Sequence(TopLevel):
    """
    The primary structure (eg, nucleotide or amino acid sequence)
    of a ComponentDefinition object.
    """

    def __init__(self, uri=URIRef("example"), elements="",
                 encoding=SBOL_ENCODING_IUPAC,
                 version=VERSION_STRING, type_uri=SBOL_SEQUENCE):
        """Construct a Sequence.

        :param uri: A full URI including a scheme, namespace, and identifier.
        If SBOLCompliance configuration is enabled, then this argument is
        simply the displayId for the new object and a full URI will
        automatically be constructed.
        :param elements: A string representation of the primary structure
        of DNA, RNA, protein, or a SMILES string for small molecules.
        :param encoding: A URI that describes the representation format
        used for the elements property. Set to SBOL_ENCODING_IUPAC by default
        :param version: An arbitrary version string. If SBOLCompliance is
        enabled, this should be a Maven version string.
        :param type_uri: The RDF type for an extension class derived
        from this one.
        """
        super().__init__(type_uri, uri, version)

        # The elements property is a REQUIRED String of characters that represents
        # the constituents of a biological or chemical molecule. For example,
        # these characters could represent the nucleotide bases
        # of a molecule of DNA, the amino acid residues of a protein,
        # or the atoms and chemical bonds of a small molecule.
        self.elements = TextProperty(self, SBOL_ELEMENTS,
                                     '1', '1', [], elements)
        self.encoding = URIProperty(self, SBOL_ENCODING,
                                    '1', '1', [], encoding)

    # The encoding property is REQUIRED and has a data type of URI.
    # This property MUST indicate how the elements property of a Sequence
    # MUST be formed and interpreted. For example, the elements property
    # of a Sequence with an IUPAC DNA encoding property MUST contain characters
    # that represent nucleotide bases, such as a, t, c, and g.
    # The elements property of a Sequence with a Simplified Molecular-Input
    # Line-Entry System (SMILES) encoding, on the other hand,
    # MUST contain characters that represent atoms and chemical bonds,
    # such as C, N, O, and =.
    #
    # It is RECOMMENDED that the encoding property contains a URI
    # from the table below. The terms in the table are organized
    # by the type of ComponentDefinition that typically refer to a Sequence
    # with such an encoding.
    # When the encoding of a Sequence is well described by one of the URIs
    # in the table, it MUST contain that URI.
    # | ComponentDefinition type  | Encoding       | libSBOL Symbol              | URI                                              |
    # | :------------------------ | :--------------| :-------------------------- | :----------------------------------------------- |
    # | DNA, RNA                  | IUPAC DNA, RNA | SBOL_ENCODING_IUPAC         | http://www.chem.qmul.ac.uk/iubmb/misc/naseq.html |
    # | Protein                   | IUPAC Protein  | SBOL_ENCODING_IUPAC_PROTEIN | http://www.chem.qmul.ac.uk/iupac/AminoAcid/      |
    # | Small Molecule            | SMILES         | SBOL_ENCODING_SMILES        | http://www.opensmiles.org/opensmiles.html        |

    def assemble(self, composite_sequence=""):
        """Calculates the complete sequence of a high-level Component
        from the sequence of its subcomponents.

        Prior to assembling the the complete sequence, you must assemble
        a template design by calling ComponentDefinition::assemble for the
        ComponentDefinition that references this Sequence.

        :param composite_sequence: Typically no value for the composite
        sequence should be specified by the user.
        This parameter is used to hold the composite sequence
        as it is passed to function calls at a higher-level of the
        recursion stack.
        :return: The complete sequence of a high-level Component.
        """
        raise NotImplementedError('Not yet implemented')

    def compile(self):
        """Synonymous with Sequence::assemble.

        Calculates the complete sequence of a high-level Component
        from the sequence of its subcomponents.
        Prior to assembling the the complete sequence, you must assemble
        a template design by calling ComponentDefinition::assemble for the
        ComponentDefinition that references this Sequence.

        :return:
        """
        self.assemble()
        return self.elements

    def __len__(self):
        """

        :return: The length of the primary sequence in the elements property.
        """
        return len(self.elements)

    def __bool__(self):
        # Ignore list semantics for boolean conversion.
        # A sequence is True if it exists.
        return True

    @deprecated(reason='Use len(sequence) instead')
    def length(self):
        """

        :return: The length of the primary sequence in the elements property.
        """
        return len(self)

    def synthesize(self, clone_id):
        """

        :param clone_id: A URI for the build, or displayId if working in
        SBOLCompliant mode.
        :return:
        """
        raise NotImplementedError("Not yet implemented")

    def compile(self, composite_sequence='', assembly_method=None):
        if not self.doc:
            raise ValueError('Cannot compile Sequence <%s>. The Sequence must belong to '
                             'a Document in order to compile.' % self.identity)

        # Search for the parent ComponentDefinition to which this Sequence belongs
        parent_cdef = None
        for cd in self.doc.componentDefinitions:
            if cd.sequence and cd.sequence.identity == self.identity:
                parent_cdef = cd
                break

        if not parent_cdef:
            raise ValueError('Cannot compile Sequence <%s>. The Sequence must be '
                             'associated with a ComponentDefinition in order to compile.'
                             % self.identity)

        if len(parent_cdef.components) == 0:
            if parent_cdef.sequence.elements:
                return parent_cdef.sequence.elements
            else:
                return ''  # Maybe this should raise an Exception ?

        elif len(parent_cdef.components) > 0:
            # Recurse into subcomponents and assemble their sequence
            composite_sequence_initial_size = len(composite_sequence)

            subcomponents = parent_cdef.getPrimaryStructureComponents()
            for c in subcomponents:
                cdef = self.doc.getComponentDefinition(c.definition)
                if not cdef.sequence:
                    if Config.getOption('sbol_compliant_uris'):
                        seq = self.doc.sequences.create(cdef.displayId)
                        # cdef.sequence = seq
                        cdef.sequences = seq.identity
                    else:
                        seq = self.doc.sequences.create(cdef.identity + '_seq')
                        # cdef.sequence = seq
                        cdef.sequences = seq.identity
                seq = cdef.sequence

                # Check for regularity -- only one SequenceAnnotation per Component is
                # allowed
                sequence_annotations = \
                    parent_cdef.find_property_value(SBOL_COMPONENT_PROPERTY, c.identity)

                if len(sequence_annotations) > 1:
                    raise SBOLError('Cannot compile Sequence. Component <%s> is '
                                    'irregular. More than one SequenceAnnotation is '
                                    'associated with this Component' % c.identity,
                                    SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT)

                # Auto-construct a SequenceAnnotation for this Component if one doesn't
                # already exist
                if len(sequence_annotations) == 0:
                    sa_instance = 0
                    sa_id = cdef.displayId if Config.getOption('sbol_compliant_uris') \
                        else cdef.identity
                    sa_id += '_annotation'
                    sa_uri = '%s/%s_%d/%s' % (parent_cdef.persistentIdentity, sa_id,
                                              sa_instance, cdef.version)
                    while sa_uri in parent_cdef.sequenceAnnotations:
                        sa_instance += 1
                        sa_uri = '%s/%s_%d/%s' % (parent_cdef.persistentIdentity, sa_id,
                                                  sa_instance, cdef.version)
                    sa = parent_cdef.sequenceAnnotations.create('%s_%d' % (sa_id,
                                                                sa_instance))
                    sa.component = c
                    sequence_annotations.append(sa)

                sa = sequence_annotations[0]

                # Check for regularity--only one Range per SequenceAnnotation is allowed
                ranges = []

                # Look for an existing Range that can be re-used
                for l in sa.locations:
                    if type(l) is Range:
                        ranges.append(l)
                else:
                    # Auto-construct a Range
                    range_id = sa.displayId if Config.getOption('sbol_compliant_uris') \
                        else sa.identity
                    r = sa.locations.createRange(range_id + '_range')
                    ranges.append(r)
                if len(ranges) > 1:
                    raise SBOLError('Cannot compile Sequence <%s> because '
                                    'SequenceAnnotation <%s> has more than one Range.'
                                    % (self.identity, sa.identity),
                                    SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT)

                r = ranges[0]
                r.start = len(composite_sequence) + 1
                subsequence = seq.compile(composite_sequence)  # Recursive call
                if assembly_method:
                    subsequence = assembly_method(subsequence)
                    if not type(subsequence) is str:
                        raise TypeError('Invalid callback specified for assembly_method.'
                                        ' The callback must return a string.')

                # If sourceLocation is specified, don't use the entire sequence for the
                # subcomponent
                if len(c.sourceLocations) == 1:
                    source_loc = c.sourceLocations.getRange()
                    subsequence = subsequence[(source_loc.start - 1):(source_loc.end)]
                composite_sequence = composite_sequence + subsequence
                r.end = len(composite_sequence)

            composite_sequence_current_size = composite_sequence_initial_size \
                + len(composite_sequence)
            subsequence = composite_sequence[composite_sequence_initial_size:
                                             composite_sequence_current_size]
            self.elements = subsequence
            return subsequence
